# ----------------------------------------------------------------------------
# pymunk
# Copyright (c) 2007-2017 Victor Blomqvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ----------------------------------------------------------------------------

"""A constraint is something that describes how two bodies interact with
each other. (how they constrain each other). Constraints can be simple
joints that allow bodies to pivot around each other like the bones in your
body, or they can be more abstract like the gear joint or motors.

This submodule contain all the constraints that are supported by Pymunk.

All the constraints support copy and pickle from the standard library. Custom 
properties set on a constraint will also be copied/pickled.

Chipmunk has a good overview of the different constraint on youtube which
works fine to showcase them in Pymunk as well.
http://www.youtube.com/watch?v=ZgJJZTS0aMM

.. raw:: html

    <iframe width="420" height="315" style="display: block; margin: 0 auto;"
    src="http://www.youtube.com/embed/ZgJJZTS0aMM" frameborder="0"
    allowfullscreen></iframe>


Example::

>>> import pymunk
>>> import pymunk.constraints
>>> s = pymunk.Space()
>>> a,b = pymunk.Body(10,10), pymunk.Body(10,10)
>>> c = pymunk.constraints.PivotJoint(a, b, (0,0))
>>> s.add(c)

"""
__docformat__ = "reStructuredText"

__all__ = [
    "Constraint",
    "PinJoint",
    "SlideJoint",
    "PivotJoint",
    "GrooveJoint",
    "DampedSpring",
    "DampedRotarySpring",
    "RotaryLimitJoint",
    "RatchetJoint",
    "GearJoint",
    "SimpleMotor",
]

import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

if TYPE_CHECKING:
    from .body import Body
    from .space import Space

from ._chipmunk_cffi import ffi, lib
from ._pickle import PickleMixin
from ._typing_attr import TypingAttrMixing
from .vec2d import Vec2d


class Constraint(PickleMixin, TypingAttrMixing, object):
    """Base class of all constraints.

    You usually don't want to create instances of this class directly, but
    instead use one of the specific constraints such as the PinJoint.
    """

    _pickle_attrs_init = PickleMixin._pickle_attrs_init + ["a", "b"]
    _pickle_attrs_general = PickleMixin._pickle_attrs_general + [
        "max_force",
        "error_bias",
        "max_bias",
        "collide_bodies",
    ]
    _pickle_attrs_skip = PickleMixin._pickle_attrs_skip + ["pre_solve", "post_solve"]

    _pre_solve_func: Optional[Callable[["Constraint", "Space"], None]] = None
    _cp_pre_solve_func: Any = ffi.NULL
    _post_solve_func: Optional[Callable[["Constraint", "Space"], None]] = None
    _cp_post_solve_func: Any = ffi.NULL

    def __init__(self, constraint: ffi.CData) -> None:
        self._constraint = constraint

    def _init(self, a: "Body", b: "Body", _constraint: Any) -> None:
        def constraintfree(cp_constraint: ffi.CData) -> None:
            cp_space = lib.cpConstraintGetSpace(cp_constraint)
            if cp_space != ffi.NULL:
                lib.cpSpaceRemoveConstraint(cp_space, cp_constraint)

            logging.debug("constraintfree %s", cp_constraint)
            lib.cpConstraintFree(cp_constraint)

        self._constraint = ffi.gc(_constraint, constraintfree)
        self._set_bodies(a, b)

    def _get_max_force(self) -> float:
        return lib.cpConstraintGetMaxForce(self._constraint)

    def _set_max_force(self, f: float) -> None:
        lib.cpConstraintSetMaxForce(self._constraint, f)

    max_force = property(
        _get_max_force,
        _set_max_force,
        doc="""The maximum force that the constraint can use to act on the two
        bodies.

        Defaults to infinity
        """,
    )

    def _get_error_bias(self) -> float:
        return lib.cpConstraintGetErrorBias(self._constraint)

    def _set_error_bias(self, error_bias: float) -> None:
        lib.cpConstraintSetErrorBias(self._constraint, error_bias)

    error_bias = property(
        _get_error_bias,
        _set_error_bias,
        doc="""The percentage of joint error that remains unfixed after a
        second.

        This works exactly the same as the collision bias property of a space,
        but applies to fixing error (stretching) of joints instead of
        overlapping collisions.

        Defaults to pow(1.0 - 0.1, 60.0) meaning that it will correct 10% of
        the error every 1/60th of a second.
        """,
    )

    def _get_max_bias(self) -> float:
        return lib.cpConstraintGetMaxBias(self._constraint)

    def _set_max_bias(self, max_bias: float) -> None:
        lib.cpConstraintSetMaxBias(self._constraint, max_bias)

    max_bias = property(
        _get_max_bias,
        _set_max_bias,
        doc="""The maximum speed at which the constraint can apply error
        correction.

        Defaults to infinity
        """,
    )

    def _get_collide_bodies(self) -> bool:
        return lib.cpConstraintGetCollideBodies(self._constraint)

    def _set_collide_bodies(self, collide_bodies: bool) -> None:
        lib.cpConstraintSetCollideBodies(self._constraint, collide_bodies)

    collide_bodies = property(
        _get_collide_bodies,
        _set_collide_bodies,
        doc="""Constraints can be used for filtering collisions too.

        When two bodies collide, Pymunk ignores the collisions if this property
        is set to False on any constraint that connects the two bodies.
        Defaults to True. This can be used to create a chain that self
        collides, but adjacent links in the chain do not collide.
        """,
    )

    @property
    def impulse(self) -> float:
        """The most recent impulse that constraint applied.

        To convert this to a force, divide by the timestep passed to
        space.step(). You can use this to implement breakable joints to check
        if the force they attempted to apply exceeded a certain threshold.
        """
        return lib.cpConstraintGetImpulse(self._constraint)

    @property
    def a(self) -> "Body":
        """The first of the two bodies constrained"""
        return self._a

    @property
    def b(self) -> "Body":
        """The second of the two bodies constrained"""
        return self._b

    def activate_bodies(self) -> None:
        """Activate the bodies this constraint is attached to"""
        self._a.activate()
        self._b.activate()

    @property
    def pre_solve(self) -> Optional[Callable[["Constraint", "Space"], None]]:
        """The pre-solve function is called before the constraint solver runs.

        Note that None can be used to reset it to default value.

        >>> import pymunk
        >>> j = pymunk.PinJoint(pymunk.Body(1,2), pymunk.Body(3,4), (0,0))
        >>> def pre_solve_func(constraint, space):
        ...     print("Hello from pre-solve")
        >>> j.pre_solve = pre_solve_func
        >>> j.pre_solve = None
        """

        return self._pre_solve_func

    @pre_solve.setter
    def pre_solve(
        self, func: Optional[Callable[["Constraint", "Space"], None]]
    ) -> None:
        self._pre_solve_func = func

        if func is None:
            self._cp_pre_solve_func = ffi.NULL
        else:

            @ffi.callback("cpConstraintPreSolveFunc")
            def _impl(cp_constraint: ffi.CData, cp_space: ffi.CData) -> None:
                assert self.a.space is not None
                func(self, self.a.space)  # type: ignore

            self._cp_pre_solve_func = _impl

        lib.cpConstraintSetPreSolveFunc(self._constraint, self._cp_pre_solve_func)

    @property
    def post_solve(self) -> Optional[Callable[["Constraint", "Space"], None]]:
        """The post-solve function is called after the constraint solver runs.

        Note that None can be used to reset it to default value.

        >>> import pymunk
        >>> j = pymunk.PinJoint(pymunk.Body(1,2), pymunk.Body(3,4), (0,0))
        >>> def post_solve_func(constraint, space):
        ...     print("Hello from pre-solve")
        >>> j.post_solve = post_solve_func
        >>> j.post_solve = None
        """
        return self._post_solve_func

    @post_solve.setter
    def post_solve(
        self, func: Optional[Callable[["Constraint", "Space"], None]]
    ) -> None:
        self._post_solve_func = func

        if func is None:
            self._cp_post_solve_func = ffi.NULL
        else:

            @ffi.callback("cpConstraintPostSolveFunc")
            def _impl(cp_constraint: ffi.CData, cp_space: ffi.CData) -> None:
                assert self.a.space is not None
                func(self, self.a.space)  # type: ignore

            self._cp_post_solve_func = _impl

        lib.cpConstraintSetPostSolveFunc(self._constraint, self._cp_post_solve_func)

    def _set_bodies(self, a: "Body", b: "Body") -> None:
        assert a is not b
        self._a = a
        self._b = b
        a._constraints.add(self)
        b._constraints.add(self)

    def __getstate__(self) -> Dict[str, List[Tuple[str, Any]]]:
        """Return the state of this object

        This method allows the usage of the :mod:`copy` and :mod:`pickle`
        modules with this class.
        """
        d = super(Constraint, self).__getstate__()

        d["special"].append(("_pre_solve_func", self._pre_solve_func))
        d["special"].append(("_post_solve_func", self._post_solve_func))

        return d

    def __setstate__(self, state: Dict[str, List[Tuple[str, Any]]]) -> None:
        """Unpack this object from a saved state.

        This method allows the usage of the :mod:`copy` and :mod:`pickle`
        modules with this class.
        """
        super(Constraint, self).__setstate__(state)

        for k, v in state["special"]:
            if k == "_pre_solve_func" and v != None:
                self.pre_solve = v
            elif k == "_post_solve_func" and v != None:
                self.post_solve = v


class PinJoint(Constraint):
    """PinJoint links shapes with a solid bar or pin.

    Keeps the anchor points at a set distance from one another.
    """

    _pickle_attrs_init = Constraint._pickle_attrs_init + ["anchor_a", "anchor_b"]

    def __init__(
        self,
        a: "Body",
        b: "Body",
        anchor_a: Tuple[float, float] = (0, 0),
        anchor_b: Tuple[float, float] = (0, 0),
    ) -> None:
        """a and b are the two bodies to connect, and anchor_a and anchor_b are
        the anchor points on those bodies.

        The distance between the two anchor points is measured when the joint
        is created. If you want to set a specific distance, use the setter
        function to override it.
        """
        assert len(anchor_a) == 2
        assert len(anchor_b) == 2
        _constraint = lib.cpPinJointNew(a._body, b._body, anchor_a, anchor_b)
        self._init(a, b, _constraint)

    def _get_anchor_a(self) -> Vec2d:
        v = lib.cpPinJointGetAnchorA(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_anchor_a(self, anchor: Tuple[float, float]) -> None:
        assert len(anchor) == 2
        lib.cpPinJointSetAnchorA(self._constraint, anchor)

    anchor_a = property(_get_anchor_a, _set_anchor_a)

    def _get_anchor_b(self) -> Vec2d:
        v = lib.cpPinJointGetAnchorB(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_anchor_b(self, anchor: Tuple[float, float]) -> None:
        assert len(anchor) == 2
        lib.cpPinJointSetAnchorB(self._constraint, anchor)

    anchor_b = property(_get_anchor_b, _set_anchor_b)

    def _get_distance(self) -> float:
        return lib.cpPinJointGetDist(self._constraint)

    def _set_distance(self, distance: float) -> None:
        lib.cpPinJointSetDist(self._constraint, distance)

    distance = property(_get_distance, _set_distance)


class SlideJoint(Constraint):
    """SlideJoint is like a PinJoint, but have a minimum and maximum distance.

    A chain could be modeled using this joint. It keeps the anchor points
    from getting to far apart, but will allow them to get closer together.
    """

    _pickle_attrs_init = Constraint._pickle_attrs_init + [
        "anchor_a",
        "anchor_b",
        "min",
        "max",
    ]

    def __init__(
        self,
        a: "Body",
        b: "Body",
        anchor_a: Tuple[float, float],
        anchor_b: Tuple[float, float],
        min: float,
        max: float,
    ) -> None:
        """a and b are the two bodies to connect, anchor_a and anchor_b are the
        anchor points on those bodies, and min and max define the allowed
        distances of the anchor points.
        """
        assert len(anchor_a) == 2
        assert len(anchor_b) == 2
        _constraint = lib.cpSlideJointNew(
            a._body, b._body, anchor_a, anchor_b, min, max
        )
        self._init(a, b, _constraint)

    def _get_anchor_a(self) -> Vec2d:
        v = lib.cpSlideJointGetAnchorA(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_anchor_a(self, anchor: Tuple[float, float]) -> None:
        assert len(anchor) == 2
        lib.cpSlideJointSetAnchorA(self._constraint, anchor)

    anchor_a = property(_get_anchor_a, _set_anchor_a)

    def _get_anchor_b(self) -> Vec2d:
        v = lib.cpSlideJointGetAnchorB(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_anchor_b(self, anchor: Tuple[float, float]) -> None:
        assert len(anchor) == 2
        lib.cpSlideJointSetAnchorB(self._constraint, anchor)

    anchor_b = property(_get_anchor_b, _set_anchor_b)

    def _get_min(self) -> float:
        return lib.cpSlideJointGetMin(self._constraint)

    def _set_min(self, min: float) -> None:
        lib.cpSlideJointSetMin(self._constraint, min)

    min = property(_get_min, _set_min)

    def _get_max(self) -> float:
        return lib.cpSlideJointGetMax(self._constraint)

    def _set_max(self, max: float) -> None:
        lib.cpSlideJointSetMax(self._constraint, max)

    max = property(_get_max, _set_max)


class PivotJoint(Constraint):
    """PivotJoint allow two objects to pivot about a single point.

    Its like a swivel.
    """

    _pickle_attrs_init = Constraint._pickle_attrs_init + ["anchor_a", "anchor_b"]

    def __init__(
        self,
        a: "Body",
        b: "Body",
        *args: Union[
            Tuple[float, float], Tuple[Tuple[float, float], Tuple[float, float]]
        ]
    ) -> None:
        """a and b are the two bodies to connect, and pivot is the point in
        world coordinates of the pivot.

        Because the pivot location is given in world coordinates, you must
        have the bodies moved into the correct positions already.
        Alternatively you can specify the joint based on a pair of anchor
        points, but make sure you have the bodies in the right place as the
        joint will fix itself as soon as you start simulating the space.

        That is, either create the joint with PivotJoint(a, b, pivot) or
        PivotJoint(a, b, anchor_a, anchor_b).

        :param Body a: The first of the two bodies
        :param Body b: The second of the two bodies
        :param args: Either one pivot point, or two anchor points
        :type args: (float,float) or (float,float) (float,float)
        """
        if len(args) == 1:
            assert len(args[0]) == 2
            _constraint = lib.cpPivotJointNew(a._body, b._body, args[0])
        elif len(args) == 2:
            assert len(args[0]) == 2
            assert len(args[1]) == 2
            _constraint = lib.cpPivotJointNew2(a._body, b._body, args[0], args[1])
        else:
            raise Exception(
                "You must specify either one pivot point" " or two anchor points"
            )

        self._init(a, b, _constraint)

    def _get_anchor_a(self) -> Vec2d:
        v = lib.cpPivotJointGetAnchorA(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_anchor_a(self, anchor: Tuple[float, float]) -> None:
        assert len(anchor) == 2
        lib.cpPivotJointSetAnchorA(self._constraint, anchor)

    anchor_a = property(_get_anchor_a, _set_anchor_a)

    def _get_anchor_b(self) -> Vec2d:
        v = lib.cpPivotJointGetAnchorB(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_anchor_b(self, anchor: Tuple[float, float]) -> None:
        assert len(anchor) == 2
        lib.cpPivotJointSetAnchorB(self._constraint, anchor)

    anchor_b = property(_get_anchor_b, _set_anchor_b)


class GrooveJoint(Constraint):
    """GrooveJoint is similar to a PivotJoint, but with a linear slide.

    One of the anchor points is a line segment that the pivot can slide in instead of being fixed.
    """

    _pickle_attrs_init = Constraint._pickle_attrs_init + [
        "groove_a",
        "groove_b",
        "anchor_b",
    ]

    def __init__(
        self,
        a: "Body",
        b: "Body",
        groove_a: Tuple[float, float],
        groove_b: Tuple[float, float],
        anchor_b: Tuple[float, float],
    ) -> None:
        """The groove goes from groove_a to groove_b on body a, and the pivot
        is attached to anchor_b on body b.

        All coordinates are body local.
        """
        assert len(groove_a) == 2
        assert len(groove_b) == 2
        assert len(anchor_b) == 2
        _constraint = lib.cpGrooveJointNew(
            a._body, b._body, groove_a, groove_b, anchor_b
        )
        self._init(a, b, _constraint)

    def _get_anchor_b(self) -> Vec2d:
        v = lib.cpGrooveJointGetAnchorB(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_anchor_b(self, anchor: Tuple[float, float]) -> None:
        assert len(anchor) == 2
        lib.cpGrooveJointSetAnchorB(self._constraint, anchor)

    anchor_b = property(_get_anchor_b, _set_anchor_b)

    def _get_groove_a(self) -> Vec2d:
        v = lib.cpGrooveJointGetGrooveA(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_groove_a(self, groove: Tuple[float, float]) -> None:
        assert len(groove) == 2
        lib.cpGrooveJointSetGrooveA(self._constraint, groove)

    groove_a = property(_get_groove_a, _set_groove_a)

    def _get_groove_b(self) -> Vec2d:
        v = lib.cpGrooveJointGetGrooveB(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_groove_b(self, groove: Tuple[float, float]) -> None:
        assert len(groove) == 2
        lib.cpGrooveJointSetGrooveB(self._constraint, groove)

    groove_b = property(_get_groove_b, _set_groove_b)


class DampedSpring(Constraint):
    """DampedSpring is a damped spring.

    The spring allows you to define the rest length, stiffness and damping.
    """

    _pickle_attrs_init = Constraint._pickle_attrs_init + [
        "anchor_a",
        "anchor_b",
        "rest_length",
        "stiffness",
        "damping",
    ]

    def __init__(
        self,
        a: "Body",
        b: "Body",
        anchor_a: Tuple[float, float],
        anchor_b: Tuple[float, float],
        rest_length: float,
        stiffness: float,
        damping: float,
    ) -> None:
        """Defined much like a slide joint.

        :param Body a: Body a
        :param Body b: Body b
        :param anchor_a: Anchor point a, relative to body a
        :type anchor_a: `(float,float)`
        :param anchor_b: Anchor point b, relative to body b
        :type anchor_b: `(float,float)`
        :param float rest_length: The distance the spring wants to be.
        :param float stiffness: The spring constant (Young's modulus).
        :param float damping: How soft to make the damping of the spring.
        """
        assert len(anchor_a) == 2
        assert len(anchor_b) == 2
        _constraint = lib.cpDampedSpringNew(
            a._body,
            b._body,
            anchor_a,
            anchor_b,
            rest_length,
            stiffness,
            damping,
        )

        self._init(a, b, _constraint)

    def _get_anchor_a(self) -> Vec2d:
        v = lib.cpDampedSpringGetAnchorA(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_anchor_a(self, anchor: Tuple[float, float]) -> None:
        assert len(anchor) == 2
        lib.cpDampedSpringSetAnchorA(self._constraint, anchor)

    anchor_a = property(_get_anchor_a, _set_anchor_a)

    def _get_anchor_b(self) -> Vec2d:
        v = lib.cpDampedSpringGetAnchorB(self._constraint)
        return Vec2d(v.x, v.y)

    def _set_anchor_b(self, anchor: Tuple[float, float]) -> None:
        assert len(anchor) == 2
        lib.cpDampedSpringSetAnchorB(self._constraint, anchor)

    anchor_b = property(_get_anchor_b, _set_anchor_b)

    def _get_rest_length(self) -> float:
        return lib.cpDampedSpringGetRestLength(self._constraint)

    def _set_rest_length(self, rest_length: float) -> None:
        lib.cpDampedSpringSetRestLength(self._constraint, rest_length)

    rest_length = property(
        _get_rest_length,
        _set_rest_length,
        doc="""The distance the spring wants to be.""",
    )

    def _get_stiffness(self) -> float:
        return lib.cpDampedSpringGetStiffness(self._constraint)

    def _set_stiffness(self, stiffness: float) -> None:
        lib.cpDampedSpringSetStiffness(self._constraint, stiffness)

    stiffness = property(
        _get_stiffness, _set_stiffness, doc="""The spring constant (Young's modulus)."""
    )

    def _get_damping(self) -> float:
        return lib.cpDampedSpringGetDamping(self._constraint)

    def _set_damping(self, damping: float) -> None:
        lib.cpDampedSpringSetDamping(self._constraint, damping)

    damping = property(
        _get_damping,
        _set_damping,
        doc="""How soft to make the damping of the spring.""",
    )


class DampedRotarySpring(Constraint):
    """DampedRotarySpring works like the DammpedSpring but in a angular fashion."""

    _pickle_attrs_init = Constraint._pickle_attrs_init + [
        "rest_angle",
        "stiffness",
        "damping",
    ]

    def __init__(
        self, a: "Body", b: "Body", rest_angle: float, stiffness: float, damping: float
    ) -> None:
        """Like a damped spring, but works in an angular fashion.

        :param Body a: Body a
        :param Body b: Body b
        :param float rest_angle: The relative angle in radians that the bodies
            want to have
        :param float stiffness: The spring constant (Young's modulus).
        :param float damping: How soft to make the damping of the spring.
        """
        _constraint = lib.cpDampedRotarySpringNew(
            a._body, b._body, rest_angle, stiffness, damping
        )
        self._init(a, b, _constraint)

    def _get_rest_angle(self) -> float:
        return lib.cpDampedRotarySpringGetRestAngle(self._constraint)

    def _set_rest_angle(self, rest_angle: float) -> None:
        lib.cpDampedRotarySpringSetRestAngle(self._constraint, rest_angle)

    rest_angle = property(
        _get_rest_angle,
        _set_rest_angle,
        doc="""The relative angle in radians that the bodies want to have""",
    )

    def _get_stiffness(self) -> float:
        return lib.cpDampedRotarySpringGetStiffness(self._constraint)

    def _set_stiffness(self, stiffness: float) -> None:
        lib.cpDampedRotarySpringSetStiffness(self._constraint, stiffness)

    stiffness = property(
        _get_stiffness, _set_stiffness, doc="""The spring constant (Young's modulus)."""
    )

    def _get_damping(self) -> float:
        return lib.cpDampedRotarySpringGetDamping(self._constraint)

    def _set_damping(self, damping: float) -> None:
        lib.cpDampedRotarySpringSetDamping(self._constraint, damping)

    damping = property(
        _get_damping,
        _set_damping,
        doc="""How soft to make the damping of the spring.""",
    )


class RotaryLimitJoint(Constraint):
    """RotaryLimitJoint constrains the relative rotations of two bodies."""

    _pickle_attrs_init = Constraint._pickle_attrs_init + ["min", "max"]

    def __init__(self, a: "Body", b: "Body", min: float, max: float) -> None:
        """Constrains the relative rotations of two bodies.

        min and max are the angular limits in radians. It is implemented so
        that it's possible to for the range to be greater than a full
        revolution.
        """
        _constraint = lib.cpRotaryLimitJointNew(a._body, b._body, min, max)
        self._init(a, b, _constraint)

    def _get_min(self) -> float:
        return lib.cpRotaryLimitJointGetMin(self._constraint)

    def _set_min(self, min: float) -> None:
        lib.cpRotaryLimitJointSetMin(self._constraint, min)

    min = property(_get_min, _set_min)

    def _get_max(self) -> float:
        return lib.cpRotaryLimitJointGetMax(self._constraint)

    def _set_max(self, max: float) -> None:
        lib.cpRotaryLimitJointSetMax(self._constraint, max)

    max = property(_get_max, _set_max)


class RatchetJoint(Constraint):
    """RatchetJoint is a rotary ratchet, it works like a socket wrench."""

    _pickle_attrs_init = Constraint._pickle_attrs_init + ["phase", "ratchet"]

    def __init__(self, a: "Body", b: "Body", phase: float, ratchet: float) -> None:
        """Works like a socket wrench.

        ratchet is the distance between "clicks", phase is the initial offset
        to use when deciding where the ratchet angles are.
        """
        _constraint = lib.cpRatchetJointNew(a._body, b._body, phase, ratchet)
        self._init(a, b, _constraint)

    def _get_angle(self) -> float:
        return lib.cpRatchetJointGetAngle(self._constraint)

    def _set_angle(self, angle: float) -> None:
        lib.cpRatchetJointSetAngle(self._constraint, angle)

    angle = property(_get_angle, _set_angle)

    def _get_phase(self) -> float:
        return lib.cpRatchetJointGetPhase(self._constraint)

    def _set_phase(self, phase: float) -> None:
        lib.cpRatchetJointSetPhase(self._constraint, phase)

    phase = property(_get_phase, _set_phase)

    def _get_ratchet(self) -> float:
        return lib.cpRatchetJointGetRatchet(self._constraint)

    def _set_ratchet(self, ratchet: float) -> None:
        lib.cpRatchetJointSetRatchet(self._constraint, ratchet)

    ratchet = property(_get_ratchet, _set_ratchet)


class GearJoint(Constraint):
    """GearJoint keeps the angular velocity ratio of a pair of bodies constant."""

    _pickle_attrs_init = Constraint._pickle_attrs_init + ["phase", "ratio"]

    def __init__(self, a: "Body", b: "Body", phase: float, ratio: float):
        """Keeps the angular velocity ratio of a pair of bodies constant.

        ratio is always measured in absolute terms. It is currently not
        possible to set the ratio in relation to a third body's angular
        velocity. phase is the initial angular offset of the two bodies.
        """
        _constraint = lib.cpGearJointNew(a._body, b._body, phase, ratio)
        self._init(a, b, _constraint)

    def _get_phase(self) -> float:
        return lib.cpGearJointGetPhase(self._constraint)

    def _set_phase(self, phase: float) -> None:
        lib.cpGearJointSetPhase(self._constraint, phase)

    phase = property(_get_phase, _set_phase)

    def _get_ratio(self) -> float:
        return lib.cpGearJointGetRatio(self._constraint)

    def _set_ratio(self, ratio: float) -> None:
        lib.cpGearJointSetRatio(self._constraint, ratio)

    ratio = property(_get_ratio, _set_ratio)


class SimpleMotor(Constraint):
    """SimpleMotor keeps the relative angular velocity constant."""

    _pickle_attrs_init = Constraint._pickle_attrs_init + ["rate"]

    def __init__(self, a: "Body", b: "Body", rate: float) -> None:
        """Keeps the relative angular velocity of a pair of bodies constant.

        rate is the desired relative angular velocity. You will usually want
        to set an force (torque) maximum for motors as otherwise they will be
        able to apply a nearly infinite torque to keep the bodies moving.
        """
        _constraint = lib.cpSimpleMotorNew(a._body, b._body, rate)
        self._init(a, b, _constraint)

    def _get_rate(self) -> float:
        return lib.cpSimpleMotorGetRate(self._constraint)

    def _set_rate(self, rate: float) -> None:
        lib.cpSimpleMotorSetRate(self._constraint, rate)

    rate = property(
        _get_rate, _set_rate, doc="""The desired relative angular velocity"""
    )



import math
from typing import NamedTuple, Tuple, Union, cast, overload

from .vec2d import Vec2d


class Transform(NamedTuple):
    """Type used for 2x3 affine transforms.

    See wikipedia for details:
    http://en.wikipedia.org/wiki/Affine_transformation

    The properties map to the matrix in this way:

    = = ==
    a c tx
    b d ty
    0 0  1
    = = ==

    An instance can be created in this way:

        >>> Transform(1,2,3,4,5,6)
        Transform(a=1, b=2, c=3, d=4, tx=5, ty=6)

    Or overriding only some of the values (on a identity matrix):

        >>> Transform(b=3,ty=5)
        Transform(a=1, b=3, c=0, d=1, tx=0, ty=5)

    Or using one of the static methods like identity or translation (see each
    method for details).

    The Transform supports the matrix multiplicaiton operator (@) with a
    Transform, Vec2d or tuple as second operand, which produces a transformed
    Transform or Vec2d as result:

        >>> Transform.scaling(2) @ Transform.scaling(3)
        Transform(a=6, b=0, c=0, d=6, tx=0, ty=0)
        >>> Transform.scaling(2) @ Vec2d(1, 2)
        Vec2d(2, 4)

    """

    a: float = 1
    b: float = 0
    c: float = 0
    d: float = 1
    tx: float = 0
    ty: float = 0

    @overload
    def __matmul__(self, other: Tuple[float, float]) -> Vec2d:
        ...

    @overload
    def __matmul__(
        self, other: Tuple[float, float, float, float, float, float]
    ) -> "Transform":
        ...

    def __matmul__(
        self,
        other: Union[
            Tuple[float, float], Tuple[float, float, float, float, float, float]
        ],
    ) -> Union[Vec2d, "Transform"]:

        """Multiply this transform with a Transform, Vec2d or Tuple of size 2
        or 6.


        Examples (Transform @ Transform):

        >>> Transform() @ Transform() == Transform.identity()
        True
        >>> Transform.translation(2,3) @ Transform.translation(4,5)
        Transform(a=1, b=0, c=0, d=1, tx=6, ty=8)
        >>> Transform.scaling(2) @ Transform.scaling(3)
        Transform(a=6, b=0, c=0, d=6, tx=0, ty=0)
        >>> Transform.scaling(2) @ Transform.translation(3,4)
        Transform(a=2, b=0, c=0, d=2, tx=6, ty=8)
        >>> Transform.translation(3,4) @ Transform.scaling(2)
        Transform(a=2, b=0, c=0, d=2, tx=3, ty=4)


        Examples (Transform @ Vec2d):

        >>> Transform.identity() @ Vec2d(1, 2)
        Vec2d(1, 2)
        >>> Transform.scaling(2) @ Vec2d(1, 2)
        Vec2d(2, 4)
        >>> Transform.translation(3,5) @ Vec2d(1, 2)
        Vec2d(4, 7)
        >>> Transform.rotation(1) @ Vec2d(1, 2) == Vec2d(1, 2).rotated(1)
        True

        """
        assert (
            len(other) == 2 or len(other) == 6
        ), f"{other} not supported. Only Vec2d, Transform and Sequence of length 2 or 6 are supported."

        if len(other) == 2:
            assert len(other) == 2
            x, y = cast(Tuple[float, float], other)
            return Vec2d(
                self.a * x + self.c * y + self.tx, self.b * x + self.d * y + self.ty
            )
        else:
            a, b, c, d, tx, ty = cast(Tuple[float, float, float, float, float, float], other)
            return Transform(
                a=self.a * a + self.c * b + self.tx * 0,
                b=self.b * a + self.d * b + self.ty * 0,
                c=self.a * c + self.c * d + self.tx * 0,
                d=self.b * c + self.d * d + self.ty * 0,
                tx=self.a * tx + self.c * ty + self.tx * 1,
                ty=self.b * tx + self.d * ty + self.ty * 1,
            )

    @staticmethod
    def identity() -> "Transform":
        """The identity transform

        Example:

        >>> Transform.identity()
        Transform(a=1, b=0, c=0, d=1, tx=0, ty=0)

        Returns a Transform with this matrix:

        = = =
        1 0 0
        0 1 0
        0 0 1
        = = =

        """
        return Transform(1, 0, 0, 1, 0, 0)

    def translated(self, x: float, y: float) -> "Transform":
        """Translate this Transform and return the result.

        Example:
        >>> Transform.scaling(2).translated(3,4)
        Transform(a=2, b=0, c=0, d=2, tx=6, ty=8)

        """
        return self @ Transform.translation(x, y)

    def scaled(self, s: float) -> "Transform":
        """Scale this Transform and return the result.

        Example:

        >>> Transform.translation(3,4).scaled(2)
        Transform(a=2, b=0, c=0, d=2, tx=3, ty=4)
        """
        return self @ Transform.scaling(s)

    def rotated(self, t: float) -> "Transform":
        """Rotate this Transform and return the result.

        >>> '%.2f, %.2f, %.2f, %.2f, %.2f, %.2f' % Transform.rotation(1).rotated(0.5)
        '0.07, 1.00, -1.00, 0.07, 0.00, 0.00'
        >>> '%.2f, %.2f, %.2f, %.2f, %.2f, %.2f' % Transform.rotation(1.5)
        '0.07, 1.00, -1.00, 0.07, 0.00, 0.00'
        """
        return self @ Transform.rotation(t)

    @staticmethod
    def translation(x: float, y: float) -> "Transform":
        """A translation transform

        Example to translate (move) by 3 on x and 5 in y axis:

        >>> Transform.translation(3, 5)
        Transform(a=1, b=0, c=0, d=1, tx=3, ty=5)

        Returns a Transform with this matrix:

        = = =
        1 0 x
        0 1 y
        0 0 1
        = = =

        """
        return Transform(tx=x, ty=y)

    # split into scale and scale_non-uniform
    @staticmethod
    def scaling(s: float) -> "Transform":
        """A scaling transform

        Example to scale 4x:

        >>> Transform.scaling(4)
        Transform(a=4, b=0, c=0, d=4, tx=0, ty=0)

        Returns a Transform with this matrix:

        = = =
        s 0 0
        0 s 0
        0 0 1
        = = =

        """
        return Transform(a=s, d=s)

    @staticmethod
    def rotation(t: float) -> "Transform":
        """A rotation transform

        Example to rotate by 1 rad:

        >>> '%.2f, %.2f, %.2f, %.2f, %.2f, %.2f' % Transform.rotation(1)
        '0.54, 0.84, -0.84, 0.54, 0.00, 0.00'

        Returns a Transform with this matrix:

        ====== ======= =
        cos(t) -sin(t) 0
        sin(t) cos(t)  0
        0      0       1
        ====== ======= =

        """
        c = math.cos(t)
        s = math.sin(t)
        return Transform(a=c, b=s, c=-s, d=c)



import argparse
import doctest
import faulthandler
import os
import os.path
import platform
import sys
import unittest
from typing import Any, Iterator

import cffi  # type: ignore

from . import doctests


def main(args: Any) -> None:
    def list_of_tests_gen(s: Any) -> Iterator[Any]:
        for test in s:
            if unittest.suite._isnotsuite(test):  # type: ignore
                yield test
            else:
                for t in list_of_tests_gen(test):
                    yield t

    path = os.path.dirname(os.path.abspath(__file__))
    suite = unittest.TestLoader().discover(path)

    doctests.load_tests(suite, args.dependencies)

    wasSuccessful = True

    filtered_suite = unittest.TestSuite()

    if args.filter is not None:
        test_filter = args.filter
        for test in list_of_tests_gen(suite):
            if (
                isinstance(test, doctest.DocTestCase)
                and test_filter.startswith("doctest")
                or test_filter in str(test.id())
            ):
                filtered_suite.addTest(test)
    else:
        filtered_suite = suite

    res = unittest.TextTestRunner(verbosity=2).run(filtered_suite)
    wasSuccessful = res.wasSuccessful()

    sys.exit(not wasSuccessful)


if sys.argv[0].endswith("__main__.py"):
    # We change sys.argv[0] to make help message more useful
    # use executable without path, unquoted
    # (it's just a hint anyway)
    # (if you have spaces in your executable you get what you deserve!)
    executable = os.path.basename(sys.executable)
    sys.argv[0] = executable + " -m pymunk.tests"


desc = """
Run the Pymunk unittests.
Note that by default it does not run with external dependencies. 
See the -d flag for details.
"""

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument(
        "-d",
        "--dependencies",
        choices=["pygame", "pyglet", "matplotlib"],
        nargs="*",
        help="Include tests with these dependencies",
        default=[],
    )
    parser.add_argument("-f", "--filter", help="Only run tests matching the filter")
    args = parser.parse_args()

    faulthandler.enable()
    print("RUNNING pymunk.tests ##################")
    print("Python / platform:")
    print(sys.version)
    print(f"on {platform.system()} {platform.machine()} using cffi {cffi.__version__}")
    print("")

    print(args)
    # sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    main(args)



import copy
from typing import Any, ClassVar, Dict, List, Tuple, TypeVar

T = TypeVar("T", bound="PickleMixin")
_State = Dict[str, List[Tuple[str, Any]]]


class PickleMixin:
    """PickleMixin is used to provide base functionality for pickle/unpickle
    and copy.
    """

    _pickle_attrs_init: ClassVar[List[str]] = []
    _pickle_attrs_general: ClassVar[List[str]] = []
    _pickle_attrs_skip: ClassVar[List[str]] = []

    def __getstate__(self) -> _State:
        """Return the state of this object

        This method allows the usage of the :mod:`copy` and :mod:`pickle`
        modules with this class.
        """

        d: _State = {
            "init": [],  # arguments for init
            "general": [],  # general attributes
            "custom": [],  # custom attributes set by user
            "special": [],  # attributes needing special handling
        }
        for a in type(self)._pickle_attrs_init:
            d["init"].append((a, self.__getattribute__(a)))

        for a in type(self)._pickle_attrs_general:
            d["general"].append((a, self.__getattribute__(a)))

        for k, v in self.__dict__.items():
            if k[0] != "_":
                d["custom"].append((k, v))

        return d

    def __setstate__(self, state: _State) -> None:
        """Unpack this object from a saved state.

        This method allows the usage of the :mod:`copy` and :mod:`pickle`
        modules with this class.
        """

        init_attrs: List[str] = []

        init_args = [v for k, v in state["init"]]
        self.__init__(*init_args)  # type: ignore

        for k, v in state["general"]:
            self.__setattr__(k, v)

        for k, v in state["custom"]:
            self.__setattr__(k, v)

    def copy(self: T) -> T:
        """Create a deep copy of this object."""
        return copy.deepcopy(self)



import pickle
import unittest

import pymunk as p
from pymunk.vec2d import Vec2d


class UnitTestBB(unittest.TestCase):
    def setUp(self) -> None:
        # print "testing pymunk version " + p.version
        pass

    def testCreation(self) -> None:
        bb_empty = p.BB()

        self.assertEqual(bb_empty.left, 0)
        self.assertEqual(bb_empty.bottom, 0)
        self.assertEqual(bb_empty.right, 0)
        self.assertEqual(bb_empty.top, 0)

        bb_defined = p.BB(-10, -5, 15, 20)

        self.assertEqual(bb_defined.left, -10)
        self.assertEqual(bb_defined.bottom, -5)
        self.assertEqual(bb_defined.right, 15)
        self.assertEqual(bb_defined.top, 20)

        bb_circle = p.BB.newForCircle((3, 3), 3)
        self.assertEqual(bb_circle.left, 0)
        self.assertEqual(bb_circle.bottom, 0)
        self.assertEqual(bb_circle.right, 6)
        self.assertEqual(bb_circle.top, 6)

    def testMerge(self) -> None:
        bb1 = p.BB(0, 0, 10, 10)
        bb2 = p.BB(2, 0, 10, 10)
        bb3 = p.BB(10, 10, 15, 15)

        self.assertEqual(bb1.merge(bb2), p.BB(0, 0, 10, 10))
        self.assertEqual(bb2.merge(bb3).merge(bb1), p.BB(0, 0, 15, 15))

    def testMethods(self) -> None:
        bb1 = p.BB(0, 0, 10, 10)
        bb2 = p.BB(10, 10, 20, 20)
        bb3 = p.BB(4, 4, 5, 5)
        bb4 = p.BB(2, 0, 10, 10)

        v1 = Vec2d(1, 1)
        v2 = Vec2d(100, 3)
        self.assertTrue(bb1.intersects(bb2))
        self.assertFalse(bb3.intersects(bb2))

        self.assertTrue(bb1.intersects_segment(v1, v2))
        self.assertFalse(bb3.intersects_segment(v1, v2))

        self.assertTrue(bb1.contains(bb3))
        self.assertFalse(bb1.contains(bb2))

        self.assertTrue(bb1.contains_vect(v1))
        self.assertFalse(bb1.contains_vect(v2))

        self.assertEqual(bb1.expand(v1), bb1)
        self.assertEqual(bb1.expand(-v2), p.BB(-100, -3, 10, 10))

        self.assertEqual(bb1.center(), (5, 5))
        self.assertEqual(bb1.area(), 100)

        self.assertEqual(bb1.merged_area(bb2), 400)

        self.assertEqual(bb2.segment_query(v1, v2), float("inf"))
        self.assertEqual(bb1.segment_query((-1, 1), (99, 1)), 0.01)

        self.assertEqual(bb1.clamp_vect(v2), Vec2d(10, 3))

        # self.assertEqual(bb1.wrap_vect((11,11)), (1,1))

    def testPickle(self) -> None:
        x = p.BB(4, 4, 5, 5)

        s = pickle.dumps(x, 2)
        actual = pickle.loads(s)

        self.assertEqual(x, actual)

        self.assertEqual(p.BB(*x), x)



__docformat__ = "reStructuredText"

from typing import NamedTuple, Tuple

from . import _chipmunk_cffi

lib = _chipmunk_cffi.lib
ffi = _chipmunk_cffi.ffi
from .vec2d import Vec2d


class BB(NamedTuple):
    """Simple axis-aligned 2D bounding box.

    Stored as left, bottom, right, top values.

    An instance can be created in this way:
        >>> BB(left=1, bottom=5, right=20, top=10)
        BB(left=1, bottom=5, right=20, top=10)

    Or partially, for example like this:
        >>> BB(right=5, top=10)
        BB(left=0, bottom=0, right=5, top=10)
    """

    left: float = 0
    bottom: float = 0
    right: float = 0
    top: float = 0

    @staticmethod
    def newForCircle(p: Tuple[float, float], r: float) -> "BB":
        """Convenience constructor for making a BB fitting a circle at
        position p with radius r.
        """

        bb_ = lib.cpBBNewForCircle(p, r)
        return BB(bb_.l, bb_.b, bb_.r, bb_.t)

    def intersects(self, other: "BB") -> bool:
        """Returns true if the bounding boxes intersect"""
        return bool(lib.cpBBIntersects(self, other))

    def intersects_segment(
        self, a: Tuple[float, float], b: Tuple[float, float]
    ) -> bool:
        """Returns true if the segment defined by endpoints a and b
        intersect this bb."""
        assert len(a) == 2
        assert len(b) == 2
        return bool(lib.cpBBIntersectsSegment(self, a, b))

    def contains(self, other: "BB") -> bool:
        """Returns true if bb completley contains the other bb"""
        return bool(lib.cpBBContainsBB(self, other))

    def contains_vect(self, v: Tuple[float, float]) -> bool:
        """Returns true if this bb contains the vector v"""
        assert len(v) == 2
        return bool(lib.cpBBContainsVect(self, v))

    def merge(self, other: "BB") -> "BB":
        """Return the minimal bounding box that contains both this bb and the
        other bb
        """
        cp_bb = lib.cpBBMerge(self, other)
        return BB(cp_bb.l, cp_bb.b, cp_bb.r, cp_bb.t)

    def expand(self, v: Tuple[float, float]) -> "BB":
        """Return the minimal bounding box that contans both this bounding box
        and the vector v
        """
        cp_bb = lib.cpBBExpand(self, tuple(v))
        return BB(cp_bb.l, cp_bb.b, cp_bb.r, cp_bb.t)

    def center(self) -> Vec2d:
        """Return the center"""
        v = lib.cpBBCenter(self)
        return Vec2d(v.x, v.y)

    def area(self) -> float:
        """Return the area"""
        return lib.cpBBArea(self)

    def merged_area(self, other: "BB") -> float:
        """Merges this and other then returns the area of the merged bounding
        box.
        """
        return lib.cpBBMergedArea(self, other)

    def segment_query(self, a: Tuple[float, float], b: Tuple[float, float]) -> float:
        """Returns the fraction along the segment query the BB is hit.

        Returns infinity if it doesnt hit
        """
        assert len(a) == 2
        assert len(b) == 2
        return lib.cpBBSegmentQuery(self, a, b)

    def clamp_vect(self, v: Tuple[float, float]) -> Vec2d:
        """Returns a copy of the vector v clamped to the bounding box"""
        assert len(v) == 2
        v2 = lib.cpBBClampVect(self, v)
        return Vec2d(v2.x, v2.y)

    '''
    def wrap_vect(self, v):
        """Returns a copy of v wrapped to the bounding box.

        That is, BB(0,0,10,10).wrap_vect((5,5)) == Vec2d._fromcffi(10,10)
        """
        return lib._cpBBWrapVect(self.cp_bb[0], v)
    '''



# ----------------------------------------------------------------------------
# pymunk
# Copyright (c) 2007-2020 Victor Blomqvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ----------------------------------------------------------------------------

"""This module contain the Vec2d class that is used in all of pymunk when a
vector is needed.

The Vec2d class is used almost everywhere in pymunk for 2d coordinates and
vectors, for example to define gravity vector in a space. However, pymunk is
smart enough to convert tuples or tuple like objects to Vec2ds so you usually
do not need to explicitly do conversions if you happen to have a tuple::

    >>> import pymunk
    >>> space = pymunk.Space()
    >>> space.gravity
    Vec2d(0.0, 0.0)
    >>> space.gravity = 3,5
    >>> space.gravity
    Vec2d(3.0, 5.0)
    >>> space.gravity += 2,6
    >>> space.gravity
    Vec2d(5.0, 11.0)

More examples::

    >>> from pymunk.vec2d import Vec2d
    >>> Vec2d(7.3, 4.2)
    Vec2d(7.3, 4.2)
    >>> Vec2d(7.3, 4.2) + Vec2d(1, 2)
    Vec2d(8.3, 6.2)

"""
__docformat__ = "reStructuredText"

import math
import numbers
import operator
from typing import NamedTuple, Tuple

__all__ = ["Vec2d"]


class Vec2d(NamedTuple):
    """2d vector class, supports vector and scalar operators, and also
    provides some high level functions.
    """

    x: float
    y: float

    # String representaion (for debugging)
    def __repr__(self) -> str:
        return "Vec2d(%s, %s)" % (self.x, self.y)

    # Addition
    def __add__(self, other: Tuple[float, float]) -> "Vec2d":  # type: ignore
        """Add a Vec2d with another Vec2d or Tuple of size 2

        >>> Vec2d(3,4) + Vec2d(1,2)
        Vec2d(4, 6)
        >>> Vec2d(3,4) + (1,2)
        Vec2d(4, 6)
        """
        assert (
            len(other) == 2
        ), f"{other} not supported. Only Vec2d and Sequence of length 2 is supported."

        return Vec2d(self.x + other[0], self.y + other[1])

    def __radd__(self, other: Tuple[float, float]) -> "Vec2d":
        """Add a Tuple of size 2 with a Vec2d

        >>> (1,2) + Vec2d(3,4)
        Vec2d(4, 6)
        """
        return self.__add__(other)

    # Subtraction
    def __sub__(self, other: Tuple[float, float]) -> "Vec2d":
        """Subtract a Vec2d with another Vec2d or Tuple of size 2

        >>> Vec2d(3,4) - Vec2d(1,2)
        Vec2d(2, 2)
        >>> Vec2d(3,4) - (1,2)
        Vec2d(2, 2)
        """
        return Vec2d(self.x - other[0], self.y - other[1])

    def __rsub__(self, other: Tuple[float, float]) -> "Vec2d":
        """Subtract a Tuple of size 2 with a Vec2d

        >>> (1,2) - Vec2d(3,4)
        Vec2d(-2, -2)
        """
        assert (
            len(other) == 2
        ), f"{other} not supported. Only Vec2d and Sequence of length 2 is supported."
        return Vec2d(other[0] - self.x, other[1] - self.y)

    # Multiplication
    def __mul__(self, other: float) -> "Vec2d":
        """Multiply with a float

        >>> Vec2d(3,6) * 2.5
        Vec2d(7.5, 15.0)
        """
        assert isinstance(other, numbers.Real)
        return Vec2d(self.x * other, self.y * other)

    def __rmul__(self, other: float) -> "Vec2d":
        """Multiply a float with a Vec2d

        >>> 2.5 * Vec2d(3,6)
        Vec2d(7.5, 15.0)
        """
        return self.__mul__(other)

    # Division
    def __floordiv__(self, other: float) -> "Vec2d":
        """Floor division by a float (also known as integer division)

        >>> Vec2d(3,6) // 2.0
        Vec2d(1.0, 3.0)
        """
        assert isinstance(other, numbers.Real)
        return Vec2d(self.x // other, self.y // other)

    def __truediv__(self, other: float) -> "Vec2d":
        """Division by a float

        >>> Vec2d(3,6) / 2.0
        Vec2d(1.5, 3.0)
        """
        assert isinstance(other, numbers.Real)
        return Vec2d(self.x / other, self.y / other)

    # Unary operations
    def __neg__(self) -> "Vec2d":
        """Return the negated version of the Vec2d

        >>> -Vec2d(1,-2)
        Vec2d(-1, 2)
        """
        return Vec2d(operator.neg(self.x), operator.neg(self.y))

    def __pos__(self) -> "Vec2d":
        """Return the unary pos of the Vec2d.

        >>> +Vec2d(1,-2)
        Vec2d(1, -2)
        """
        return Vec2d(operator.pos(self.x), operator.pos(self.y))

    def __abs__(self) -> float:
        """Return the length of the Vec2d

        >>> abs(Vec2d(3,4))
        5.0
        """
        return self.length

    # vectory functions
    def get_length_sqrd(self) -> float:
        """Get the squared length of the vector.
        If the squared length is enough it is more efficient to use this method
        instead of first calling get_length() or access .length and then do a
        x**2.

        >>> v = Vec2d(3,4)
        >>> v.get_length_sqrd() == v.length**2
        True

        :return: The squared length
        """
        return self.x ** 2 + self.y ** 2

    @property
    def length(self) -> float:
        """Get the length of the vector.

        >>> Vec2d(10, 0).length
        10.0
        >>> '%.2f' % Vec2d(10, 20).length
        '22.36'

        :return: The length
        """
        return math.sqrt(self.x ** 2 + self.y ** 2)

    def scale_to_length(self, length: float) -> "Vec2d":
        """Return a copy of this vector scaled to the given length.

        >>> '%.2f, %.2f' % Vec2d(10, 20).scale_to_length(20)
        '8.94, 17.89'
        """
        old_length = self.length
        return Vec2d(self.x * length / old_length, self.y * length / old_length)

    def rotated(self, angle_radians: float) -> "Vec2d":
        """Create and return a new vector by rotating this vector by
        angle_radians radians.

        :return: Rotated vector
        """
        cos = math.cos(angle_radians)
        sin = math.sin(angle_radians)
        x = self.x * cos - self.y * sin
        y = self.x * sin + self.y * cos
        return Vec2d(x, y)

    def rotated_degrees(self, angle_degrees: float) -> "Vec2d":
        """Create and return a new vector by rotating this vector by
        angle_degrees degrees.

        :return: Rotade vector
        """
        return self.rotated(math.radians(angle_degrees))

    @property
    def angle(self) -> float:
        """The angle (in radians) of the vector"""
        if self.get_length_sqrd() == 0:
            return 0
        return math.atan2(self.y, self.x)

    @property
    def angle_degrees(self) -> float:
        """Gets the angle (in degrees) of a vector"""
        return math.degrees(self.angle)

    def get_angle_between(self, other: Tuple[float, float]) -> float:
        """Get the angle between the vector and the other in radians

        :return: The angle
        """
        assert len(other) == 2
        cross = self.x * other[1] - self.y * other[0]
        dot = self.x * other[0] + self.y * other[1]
        return math.atan2(cross, dot)

    def get_angle_degrees_between(self, other: "Vec2d") -> float:
        """Get the angle between the vector and the other in degrees

        :return: The angle (in degrees)
        """
        return math.degrees(self.get_angle_between(other))

    def normalized(self) -> "Vec2d":
        """Get a normalized copy of the vector
        Note: This function will return 0 if the length of the vector is 0.

        :return: A normalized vector
        """
        length = self.length
        if length != 0:
            return self / length
        return Vec2d(0, 0)

    def normalized_and_length(self) -> Tuple["Vec2d", float]:
        """Normalize the vector and return its length before the normalization

        :return: The length before the normalization
        """
        length = self.length
        if length != 0:
            return self / length, length
        return Vec2d(0, 0), 0

    def perpendicular(self) -> "Vec2d":
        return Vec2d(-self.y, self.x)

    def perpendicular_normal(self) -> "Vec2d":
        length = self.length
        if length != 0:
            return Vec2d(-self.y / length, self.x / length)
        return Vec2d(self.x, self.y)

    def dot(self, other: Tuple[float, float]) -> float:
        """The dot product between the vector and other vector
            v1.dot(v2) -> v1.x*v2.x + v1.y*v2.y

        :return: The dot product
        """
        assert len(other) == 2
        return float(self.x * other[0] + self.y * other[1])

    def get_distance(self, other: Tuple[float, float]) -> float:
        """The distance between the vector and other vector

        :return: The distance
        """
        assert len(other) == 2
        return math.sqrt((self.x - other[0]) ** 2 + (self.y - other[1]) ** 2)

    def get_dist_sqrd(self, other: Tuple[float, float]) -> float:
        """The squared distance between the vector and other vector
        It is more efficent to use this method than to call get_distance()
        first and then do a sqrt() on the result.

        :return: The squared distance
        """
        assert len(other) == 2
        return (self.x - other[0]) ** 2 + (self.y - other[1]) ** 2

    def projection(self, other: Tuple[float, float]) -> "Vec2d":
        """Project this vector on top of other vector"""
        assert len(other) == 2
        other_length_sqrd = other[0] * other[0] + other[1] * other[1]
        if other_length_sqrd == 0.0:
            return Vec2d(0, 0)
        projected_length_times_other_length = self.dot(other)
        new_length = projected_length_times_other_length / other_length_sqrd
        return Vec2d(other[0] * new_length, other[1] * new_length)

    def cross(self, other: Tuple[float, float]) -> float:
        """The cross product between the vector and other vector
            v1.cross(v2) -> v1.x*v2.y - v1.y*v2.x

        :return: The cross product
        """
        assert len(other) == 2
        return self.x * other[1] - self.y * other[0]

    def interpolate_to(self, other: Tuple[float, float], range: float) -> "Vec2d":
        assert len(other) == 2
        return Vec2d(
            self.x + (other[0] - self.x) * range, self.y + (other[1] - self.y) * range
        )

    def convert_to_basis(
        self, x_vector: Tuple[float, float], y_vector: Tuple[float, float]
    ) -> "Vec2d":
        assert len(x_vector) == 2
        assert len(y_vector) == 2
        x = self.dot(x_vector) / Vec2d(*x_vector).get_length_sqrd()
        y = self.dot(y_vector) / Vec2d(*y_vector).get_length_sqrd()
        return Vec2d(x, y)

    @property
    def int_tuple(self) -> Tuple[int, int]:
        """The x and y values of this vector as a tuple of ints.
        Uses round() to round to closest int.

        >>> Vec2d(0.9, 2.4).int_tuple
        (1, 2)
        """
        return round(self.x), round(self.y)

    @staticmethod
    def zero() -> "Vec2d":
        """A vector of zero length.

        >>> Vec2d.zero()
        Vec2d(0, 0)
        """
        return Vec2d(0, 0)

    @staticmethod
    def unit() -> "Vec2d":
        """A unit vector pointing up

        >>> Vec2d.unit()
        Vec2d(0, 1)
        """
        return Vec2d(0, 1)

    @staticmethod
    def ones() -> "Vec2d":
        """A vector where both x and y is 1

        >>> Vec2d.ones()
        Vec2d(1, 1)
        """
        return Vec2d(1, 1)

    # Extra functions, mainly for chipmunk
    def cpvrotate(self, other: Tuple[float, float]) -> "Vec2d":
        """Uses complex multiplication to rotate this vector by the other. """
        assert len(other) == 2
        return Vec2d(
            self.x * other[0] - self.y * other[1], self.x * other[1] + self.y * other[0]
        )

    def cpvunrotate(self, other: Tuple[float, float]) -> "Vec2d":
        """The inverse of cpvrotate"""
        assert len(other) == 2
        return Vec2d(
            self.x * other[0] + self.y * other[1], self.y * other[0] - self.x * other[1]
        )



"""This submodule contains helper functions to help with quick prototyping 
using pymunk together with pyglet.

Intended to help with debugging and prototyping, not for actual production use
in a full application. The methods contained in this module is opinionated 
about your coordinate system and not very optimized (they use batched 
drawing, but there is probably room for optimizations still). 
"""

__docformat__ = "reStructuredText"

from typing import TYPE_CHECKING, Any, Sequence

import matplotlib.pyplot as plt  # type: ignore

import pymunk
from pymunk.space_debug_draw_options import SpaceDebugColor
from pymunk.vec2d import Vec2d

if TYPE_CHECKING:
    import matplotlib as mpl  # type: ignore


class DrawOptions(pymunk.SpaceDebugDrawOptions):
    def __init__(self, ax: Any) -> None:
        """DrawOptions for space.debug_draw() to draw a space on a ax object.

        Typical usage::

        >>> import matplotlib as mpl
        >>> import matplotlib.pyplot as plt
        >>> import pymunk
        >>> import pymunk.matplotlib_util
        >>> space = pymunk.Space()
        >>> ax = plt.subplot()
        >>> options = pymunk.matplotlib_util.DrawOptions(ax)
        >>> space.debug_draw(options)

        You can control the color of a Shape by setting shape.color to the color
        you want it drawn in.

        >>> shape = pymunk.Circle(space.static_body, 10)
        >>> shape.color = (1, 0, 0, 1) # will draw shape in red

        See matplotlib_util.demo.py for a full example

        :Param:
            ax: matplotlib.Axes
                A matplotlib Axes object.

        """
        super(DrawOptions, self).__init__()

        self.ax = ax

    def draw_circle(
        self,
        pos: Vec2d,
        angle: float,
        radius: float,
        outline_color: SpaceDebugColor,
        fill_color: SpaceDebugColor,
    ) -> None:
        p = plt.Circle(  # type: ignore
            pos,
            radius,
            facecolor=fill_color.as_float(),
            edgecolor=outline_color.as_float(),
        )
        self.ax.add_patch(p)

        circle_edge = pos + Vec2d(radius, 0).rotated(angle)
        line = plt.Line2D(  # type: ignore
            [pos.x, circle_edge.x],
            [pos.y, circle_edge.y],
            linewidth=1,
            color=outline_color.as_float(),
        )
        line.set_solid_capstyle("round")  # type: ignore
        self.ax.add_line(line)

    def draw_segment(self, a: Vec2d, b: Vec2d, color: SpaceDebugColor) -> None:
        line = plt.Line2D([a.x, b.x], [a.y, b.y], linewidth=1, color=color.as_float())  # type: ignore
        line.set_solid_capstyle("round")  # type: ignore
        self.ax.add_line(line)

    def draw_fat_segment(
        self,
        a: Vec2d,
        b: Vec2d,
        radius: float,
        outline_color: SpaceDebugColor,
        fill_color: SpaceDebugColor,
    ) -> None:
        radius = max(1, 2 * radius)
        line = plt.Line2D(  # type: ignore
            [a.x, b.x], [a.y, b.y], linewidth=radius, color=fill_color.as_float()
        )
        line.set_solid_capstyle("round")  # type: ignore
        self.ax.add_line(line)

    def draw_polygon(
        self,
        verts: Sequence[Vec2d],
        radius: float,
        outline_color: SpaceDebugColor,
        fill_color: SpaceDebugColor,
    ) -> None:
        radius = max(1, 2 * radius)
        p = plt.Polygon(  # type: ignore
            verts,
            linewidth=radius,
            joinstyle="round",
            facecolor=fill_color.as_float(),
            edgecolor=outline_color.as_float(),
        )
        self.ax.add_patch(p)

    def draw_dot(self, size: float, pos: Vec2d, color: SpaceDebugColor) -> None:
        p = plt.Circle(pos, size, facecolor=color.as_float(), edgecolor="None")  # type: ignore
        self.ax.add_patch(p)



import pickle
import unittest

import pymunk as p


class UnitTestShapeFilter(unittest.TestCase):
    def testInit(self) -> None:
        f = p.ShapeFilter()
        self.assertEqual(f.group, 0)
        self.assertEqual(f.categories, 0xFFFFFFFF)
        self.assertEqual(f.mask, 0xFFFFFFFF)

        f = p.ShapeFilter(1, 2, 3)
        self.assertEqual(f.group, 1)
        self.assertEqual(f.categories, 2)
        self.assertEqual(f.mask, 3)

    def testConstants(self) -> None:
        self.assertEqual(p.ShapeFilter.ALL_MASKS(), 0xFFFFFFFF)
        self.assertEqual(p.ShapeFilter.ALL_CATEGORIES(), 0xFFFFFFFF)

    def testEq(self) -> None:
        f1 = p.ShapeFilter(1, 2, 3)
        f2 = p.ShapeFilter(1, 2, 3)
        f3 = p.ShapeFilter(2, 3, 4)
        self.assertTrue(f1 == f2)
        self.assertTrue(f1 != f3)

    def testPickle(self) -> None:
        x = p.ShapeFilter(1, 2, 3)
        s = pickle.dumps(x, 2)
        actual = pickle.loads(s)
        self.assertEqual(x, actual)


class UnitTestContactPoint(unittest.TestCase):
    pass


class UnitTestContactPointSet(unittest.TestCase):
    pass



import doctest
import pkgutil
import sys
import unittest
from typing import Any, List

import pymunk

ignores = ["pymunk_extension_build"]
all_dependencies = ["pygame", "pyglet", "matplotlib"]


def load_tests(
    tests: unittest.TestSuite, dependencies: List[str] = []
) -> unittest.TestSuite:
    for importer, modname, ispkg in pkgutil.iter_modules(pymunk.__path__):  # type: ignore  # mypy issue #1422
        # try:
        skip = False
        if modname in ignores:
            skip = True
        for dep in all_dependencies:
            if modname.startswith(dep) and dep not in dependencies:
                skip = True
        if skip:
            continue
        tests.addTests(doctest.DocTestSuite("pymunk." + modname))

        # except Exception as e:
        #    print("Skipping " + modname, e)
    tests.addTests(doctest.DocTestSuite(pymunk))
    return tests


if __name__ == "__main__":
    print("running doctests")
    suite = unittest.TestSuite()
    load_tests(suite)
    res = unittest.TextTestRunner().run(suite)
    sys.exit(not res.wasSuccessful())



# ----------------------------------------------------------------------------
# pymunk
# Copyright (c) 2007-2020 Victor Blomqvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ----------------------------------------------------------------------------

"""
Internal module used to make it possible to import the pymunk verison number 
from more places than __init__.

"""

from . import _chipmunk_cffi

cp = _chipmunk_cffi.lib
ffi = _chipmunk_cffi.ffi

version = "6.2.1"

chipmunk_version = "%s-%s" % (
    ffi.string(cp.cpVersionString).decode("utf-8"),
    "0593976ef47fcb3957166bd342f6b2bafe4d0e44 ",
)



from typing import Any


class TypingAttrMixing:
    """Type helper mixin to make mypy accept dynamic attribtutes.
    
    """

    def __setattr__(self, name: str, value: Any) -> None:
        """Override default setattr to make sure type checking works."""
        super().__setattr__(name, value)

    def __getattr__(self, name: str) -> Any:
        """Override default getattr to make sure type checking works."""
        return self.__getattribute__(name)



import unittest
from typing import List, Tuple

import pymunk.autogeometry as a
from pymunk.bb import BB
from pymunk.vec2d import Vec2d


class UnitTestPolylineSet(unittest.TestCase):
    def test_collect_segment(self) -> None:
        pset = a.PolylineSet()

        pset.collect_segment((0, 0), (5, 5))
        pset.collect_segment((5, 5), (10, 10))

        pset.collect_segment((2, 5), (2, 10))

        expected = [
            [Vec2d(0.0, 0.0), Vec2d(5.0, 5.0), Vec2d(10.0, 10.0)],
            [Vec2d(2.0, 5.0), Vec2d(2.0, 10.0)],
        ]

        self.assertEqual(len(pset), 2)
        self.assertEqual(list(pset), expected)


class UnitTestAutoGeometry(unittest.TestCase):
    def test_is_closed(self) -> None:
        not_closed: List[Tuple[float, float]] = [(0, 0), (1, 1), (0, 1)]
        closed: List[Tuple[float, float]] = [(0, 0), (1, 1), (0, 1), (0, 0)]
        self.assertFalse(a.is_closed(not_closed))
        self.assertTrue(a.is_closed(closed))

    def test_simplify_curves(self) -> None:
        p1: List[Tuple[float, float]] = [(0, 0), (0, 10), (5, 11), (10, 10), (0, 10)]
        expected = [(0, 0), (0, 10), (10, 10), (0, 10)]
        actual = a.simplify_curves(p1, 1)
        self.assertEqual(actual, expected)

    def test_simplify_vertexes(self) -> None:
        p1: List[Tuple[float, float]] = [(0, 0), (0, 10), (5, 11), (10, 10), (0, 10)]
        expected = [(0, 0), (0, 10), (10, 10), (0, 10)]
        actual = a.simplify_vertexes(p1, 1)
        self.assertEqual(actual, expected)

    def test_to_convex_hull(self) -> None:
        p1: List[Tuple[float, float]] = [(0, 0), (0, 10), (5, 5), (10, 10), (10, 0)]
        expected = [(0, 0), (10, 0), (10, 10), (0, 10), (0, 0)]
        actual = a.to_convex_hull(p1, 1)
        self.assertEqual(actual, expected)

    def test_convex_decomposition(self) -> None:
        # TODO: Use a more complicated polygon as test case
        p1: List[Tuple[float, float]] = [
            (0, 0),
            (5, 0),
            (10, 10),
            (20, 20),
            (5, 5),
            (0, 10),
            (0, 0),
        ]
        expected = [
            [(5.0, 5.0), (6.25, 2.5), (20.0, 20.0), (5.0, 5.0)],
            [(0.0, 0.0), (5.0, 0.0), (6.25, 2.5), (5.0, 5.0), (0.0, 10.0), (0.0, 0.0)],
        ]

        actual = a.convex_decomposition(p1, 0.1)
        actual.sort(key=len)

        # TODO: The result of convex_decomposition is not stable between
        # environments, so we cant have this assert here.
        # self.assertEqual(actual, expected)

    def test_march_soft(self) -> None:
        img = [
            "  xx   ",
            "  xx   ",
            "  xx   ",
            "  xx   ",
            "  xx   ",
            "  xxxxx",
            "  xxxxx",
        ]

        def sample_func(point: Tuple[float, float]) -> float:
            x = int(point[0])
            y = int(point[1])
            if img[y][x] == "x":
                return 1
            return 0

        pl_set = a.march_soft(BB(0, 0, 6, 6), 7, 7, 0.5, sample_func)

        expected = [
            [
                (1.5, 6.0),
                (1.5, 5.0),
                (1.5, 4.0),
                (1.5, 3.0),
                (1.5, 2.0),
                (1.5, 1.0),
                (1.5, 0.0),
            ],
            [
                (3.5, 0.0),
                (3.5, 1.0),
                (3.5, 2.0),
                (3.5, 3.0),
                (3.5, 4.0),
                (4.0, 4.5),
                (5.0, 4.5),
                (6.0, 4.5),
            ],
        ]
        self.assertEqual(list(pl_set), expected)

    def test_march_hard(self) -> None:
        img = [
            "  xx   ",
            "  xx   ",
            "  xx   ",
            "  xx   ",
            "  xx   ",
            "  xxxxx",
            "  xxxxx",
        ]

        def sample_func(point: Tuple[float, float]) -> float:
            x = int(point[0])
            y = int(point[1])
            if img[y][x] == "x":
                return 1
            return 0

        actual = list(a.march_hard(BB(0, 0, 6, 6), 7, 7, 0.5, sample_func))

        expected = [
            [
                (1.5, 6.0),
                (1.5, 5.0),
                (1.5, 4.0),
                (1.5, 3.0),
                (1.5, 2.0),
                (1.5, 1.0),
                (1.5, 0.0),
            ],
            [
                (3.5, 0.0),
                (3.5, 1.0),
                (3.5, 2.0),
                (3.5, 3.0),
                (3.5, 4.0),
                (3.5, 4.5),
                (4.0, 4.5),
                (5.0, 4.5),
                (6.0, 4.5),
            ],
        ]
        self.assertEqual(actual, expected)



import unittest
from typing import Any

import pymunk as p
from pymunk.vec2d import Vec2d

####################################################################


class UnitTestGeneral(unittest.TestCase):
    def testGeneral(self) -> None:
        p.version
        p.chipmunk_version

    def testMomentHelpers(self) -> None:
        m = p.moment_for_circle(1, 2, 3, (1, 2))
        self.assertAlmostEqual(m, 11.5)

        m = p.moment_for_segment(1, (-10, 0), (10, 0), 1)
        self.assertAlmostEqual(m, 40.6666666666)

        m = p.moment_for_poly(1, [(0, 0), (10, 10), (10, 0)], (1, 2), 3)
        self.assertAlmostEqual(m, 98.3333333333)

        m = p.moment_for_box(1, (2, 3))
        self.assertAlmostEqual(m, 1.08333333333)

    def testAreaHelpers(self) -> None:
        a = p.area_for_circle(1, 2)
        self.assertAlmostEqual(a, 9.4247779607)

        a = p.area_for_segment((-10, 0), (10, 0), 3)
        self.assertAlmostEqual(a, 148.27433388)

        a = p.area_for_poly([(0, 0), (10, 10), (10, 0)], 3)
        self.assertAlmostEqual(a, 80.700740753)


class UnitTestBugs(unittest.TestCase):
    def testManyBoxCrash(self) -> None:
        space = p.Space()
        for x in [1, 2]:
            for y in range(16):
                size = 10
                box_points = [
                    Vec2d(*p)
                    for p in [
                        (-size, -size),
                        (-size, size),
                        (size, size),
                        (size, -size),
                    ]
                ]
                body = p.Body(10, 20)
                shape = p.Poly(body, box_points)
                space.add(body, shape)
            space.step(1 / 50.0)

    def testNoStaticShape(self) -> None:
        space = p.Space()

        b1 = p.Body(1, float("inf"))
        c1 = p.Circle(b1, 10)
        c1.name = "c1"
        c1.collision_type = 2

        b2 = p.Body(1, float("inf"))
        c2 = p.Circle(b2, 10)
        c2.name = "c2"

        b3 = p.Body(1, float("inf"))
        c3 = p.Circle(b3, 10)
        c3.name = "c3"

        b1.position = 0, 0
        b2.position = 9, 0
        b3.position = -9, 0

        space.add(b1, c1, b2, c2, b3, c3)

        def remove_first(arbiter: p.Arbiter, space: p.Space, data: Any) -> None:
            first_shape = arbiter.shapes[0]
            if c1 in space.shapes:
                space.remove(c1)
            # space.add_post_step_callback(space.remove, first_shape, first_shape.body)
            space.remove(c1)

        space.add_collision_handler(2, 0).separate = remove_first

        space.step(1.0 / 60)
        b2.position = 22, 0
        space.step(1.0 / 60)


####################################################################
if __name__ == "__main__":
    print("testing pymunk version " + p.version)
    unittest.main()



__docformat__ = "reStructuredText"

from typing import TYPE_CHECKING, NamedTuple, Optional

if TYPE_CHECKING:
    from .contact_point_set import ContactPointSet
    from .shapes import Shape
    from .vec2d import Vec2d


class PointQueryInfo(NamedTuple):
    """PointQueryInfo holds the result of a point query made on a Shape or
    Space.
    """

    shape: Optional["Shape"]
    """The nearest shape, None if no shape was within range."""

    point: "Vec2d"
    """The closest point on the shape's surface. (in world space
    coordinates)
    """

    distance: float
    """The distance to the point. The distance is negative if the
    point is inside the shape.
    """

    gradient: "Vec2d"
    """The gradient of the signed distance function.

    The value should be similar to
    PointQueryInfo.point/PointQueryInfo.distance, but accurate even for
    very small values of info.distance.
    """


class SegmentQueryInfo(NamedTuple):
    """Segment queries return more information than just a simple yes or no,
    they also return where a shape was hit and it's surface normal at the hit
    point. This object hold that information.

    To test if the query hit something, check if
    SegmentQueryInfo.shape == None or not.

    Segment queries are like ray casting, but because not all spatial indexes
    allow processing infinitely long ray queries it is limited to segments.
    In practice this is still very fast and you don't need to worry too much
    about the performance as long as you aren't using extremely long segments
    for your queries.

    """

    shape: Optional["Shape"]
    """Shape that was hit, or None if no collision occured"""

    point: "Vec2d"
    """The point of impact."""

    normal: "Vec2d"
    """The normal of the surface hit."""

    alpha: float
    """The normalized distance along the query segment in the range [0, 1]"""


class ShapeQueryInfo(NamedTuple):
    """Shape queries return more information than just a simple yes or no,
    they also return where a shape was hit. This object hold that information.
    """

    shape: Optional["Shape"]
    """Shape that was hit, or None if no collision occured"""

    contact_point_set: "ContactPointSet"



__docformat__ = "reStructuredText"

import logging
from typing import TYPE_CHECKING, Any, List, Optional, Sequence, Tuple

if TYPE_CHECKING:
    from .body import Body
    from .space import Space

from ._chipmunk_cffi import ffi
from ._chipmunk_cffi import lib as cp
from ._pickle import PickleMixin, _State
from ._typing_attr import TypingAttrMixing
from .bb import BB
from .contact_point_set import ContactPointSet
from .query_info import PointQueryInfo, SegmentQueryInfo
from .shape_filter import ShapeFilter
from .transform import Transform
from .vec2d import Vec2d


class Shape(PickleMixin, TypingAttrMixing, object):
    """Base class for all the shapes.

    You usually dont want to create instances of this class directly but use
    one of the specialized shapes instead (:py:class:`Circle`,
    :py:class:`Poly` or :py:class:`Segment`).

    All the shapes can be copied and pickled. If you copy/pickle a shape the
    body (if any) will also be copied.
    """

    _pickle_attrs_init = PickleMixin._pickle_attrs_init + ["body"]
    _pickle_attrs_general = PickleMixin._pickle_attrs_general + [
        "sensor",
        "collision_type",
        "filter",
        "elasticity",
        "friction",
        "surface_velocity",
    ]
    _pickle_attrs_skip = PickleMixin._pickle_attrs_skip + ["mass", "density"]

    _space = None  # Weak ref to the space holding this body (if any)

    _id_counter = 1

    def __init__(self, shape: "Shape") -> None:
        self._shape = shape
        self._body = shape.body

    def _init(self, body: Optional["Body"], _shape: ffi.CData) -> None:
        self._body = body

        if body is not None:
            body._shapes.add(self)

        def shapefree(cp_shape):  # type: ignore
            cp_space = cp.cpShapeGetSpace(cp_shape)
            if cp_space != ffi.NULL:
                logging.debug("shapefree remove from space %s %s", cp_space, cp_shape)
                cp.cpSpaceRemoveShape(cp_space, cp_shape)

            logging.debug("shapefree remove body %s", cp_shape)
            cp.cpShapeSetBody(cp_shape, ffi.NULL)
            logging.debug("shapefree free %s", cp_shape)
            cp.cpShapeFree(cp_shape)

        self._shape = ffi.gc(_shape, shapefree)
        self._set_id()

    @property
    def _id(self) -> int:
        """Unique id of the Shape

        .. note::
            Experimental API. Likely to change in future major, minor orpoint
            releases.
        """
        return int(ffi.cast("int", cp.cpShapeGetUserData(self._shape)))

    def _set_id(self) -> None:
        cp.cpShapeSetUserData(self._shape, ffi.cast("cpDataPointer", Shape._id_counter))
        Shape._id_counter += 1

    def _get_mass(self) -> float:
        return cp.cpShapeGetMass(self._shape)

    def _set_mass(self, mass: float) -> None:
        cp.cpShapeSetMass(self._shape, mass)

    mass = property(
        _get_mass,
        _set_mass,
        doc="""The mass of this shape.

        This is useful when you let Pymunk calculate the total mass and inertia 
        of a body from the shapes attached to it. (Instead of setting the body 
        mass and inertia directly)
        """,
    )

    def _get_density(self) -> float:
        return cp.cpShapeGetDensity(self._shape)

    def _set_density(self, density: float) -> None:
        cp.cpShapeSetDensity(self._shape, density)

    density = property(
        _get_density,
        _set_density,
        doc="""The density of this shape.
        
        This is useful when you let Pymunk calculate the total mass and inertia 
        of a body from the shapes attached to it. (Instead of setting the body 
        mass and inertia directly)
        """,
    )

    @property
    def moment(self) -> float:
        """The calculated moment of this shape."""
        return cp.cpShapeGetMoment(self._shape)

    @property
    def area(self) -> float:
        """The calculated area of this shape."""
        return cp.cpShapeGetArea(self._shape)

    @property
    def center_of_gravity(self) -> Vec2d:
        """The calculated center of gravity of this shape."""
        v = cp.cpShapeGetCenterOfGravity(self._shape)
        return Vec2d(v.x, v.y)

    def _get_sensor(self) -> bool:
        return bool(cp.cpShapeGetSensor(self._shape))

    def _set_sensor(self, is_sensor: bool) -> None:
        cp.cpShapeSetSensor(self._shape, is_sensor)

    sensor = property(
        _get_sensor,
        _set_sensor,
        doc="""A boolean value if this shape is a sensor or not.

        Sensors only call collision callbacks, and never generate real
        collisions.
        """,
    )

    def _get_collision_type(self) -> int:
        return cp.cpShapeGetCollisionType(self._shape)

    def _set_collision_type(self, t: int) -> None:
        cp.cpShapeSetCollisionType(self._shape, t)

    collision_type = property(
        _get_collision_type,
        _set_collision_type,
        doc="""User defined collision type for the shape.

        See :py:meth:`Space.add_collision_handler` function for more 
        information on when to use this property.
        """,
    )

    def _get_filter(self) -> ShapeFilter:
        f = cp.cpShapeGetFilter(self._shape)
        return ShapeFilter(f.group, f.categories, f.mask)

    def _set_filter(self, f: ShapeFilter) -> None:
        cp.cpShapeSetFilter(self._shape, f)

    filter = property(
        _get_filter,
        _set_filter,
        doc="""Set the collision :py:class:`ShapeFilter` for this shape.
        """,
    )

    def _get_elasticity(self) -> float:
        return cp.cpShapeGetElasticity(self._shape)

    def _set_elasticity(self, e: float) -> None:
        cp.cpShapeSetElasticity(self._shape, e)

    elasticity = property(
        _get_elasticity,
        _set_elasticity,
        doc="""Elasticity of the shape.

        A value of 0.0 gives no bounce, while a value of 1.0 will give a
        'perfect' bounce. However due to inaccuracies in the simulation
        using 1.0 or greater is not recommended.
        """,
    )

    def _get_friction(self) -> float:
        return cp.cpShapeGetFriction(self._shape)

    def _set_friction(self, u: float) -> None:
        cp.cpShapeSetFriction(self._shape, u)

    friction = property(
        _get_friction,
        _set_friction,
        doc="""Friction coefficient.

        Pymunk uses the Coulomb friction model, a value of 0.0 is
        frictionless.

        A value over 1.0 is perfectly fine.

        Some real world example values from Wikipedia (Remember that
        it is what looks good that is important, not the exact value).

        ==============  ======  ========
        Material        Other   Friction
        ==============  ======  ========
        Aluminium       Steel   0.61
        Copper          Steel   0.53
        Brass           Steel   0.51
        Cast iron       Copper  1.05
        Cast iron       Zinc    0.85
        Concrete (wet)  Rubber  0.30
        Concrete (dry)  Rubber  1.0
        Concrete        Wood    0.62
        Copper          Glass   0.68
        Glass           Glass   0.94
        Metal           Wood    0.5
        Polyethene      Steel   0.2
        Steel           Steel   0.80
        Steel           Teflon  0.04
        Teflon (PTFE)   Teflon  0.04
        Wood            Wood    0.4
        ==============  ======  ========
        """,
    )

    def _get_surface_velocity(self) -> Vec2d:
        v = cp.cpShapeGetSurfaceVelocity(self._shape)
        return Vec2d(v.x, v.y)

    def _set_surface_velocity(self, surface_v: Vec2d) -> None:
        assert len(surface_v) == 2
        cp.cpShapeSetSurfaceVelocity(self._shape, surface_v)

    surface_velocity = property(
        _get_surface_velocity,
        _set_surface_velocity,
        doc="""The surface velocity of the object.

        Useful for creating conveyor belts or players that move around. This
        value is only used when calculating friction, not resolving the
        collision.
        """,
    )

    def _get_body(self) -> Optional["Body"]:
        return self._body

    def _set_body(self, body: Optional["Body"]) -> None:
        if self._body != None:
            self._body._shapes.remove(self)
        body_body = ffi.NULL if body is None else body._body
        cp.cpShapeSetBody(self._shape, body_body)
        if body is not None:
            body._shapes.add(self)
        self._body = body

    body = property(
        _get_body,
        _set_body,
        doc="""The body this shape is attached to. Can be set to None to
        indicate that this shape doesnt belong to a body.""",
    )

    def update(self, transform: Transform) -> BB:
        """Update, cache and return the bounding box of a shape with an
        explicit transformation.

        Useful if you have a shape without a body and want to use it for
        querying.
        """
        _bb = cp.cpShapeUpdate(self._shape, transform)
        return BB(_bb.l, _bb.b, _bb.r, _bb.t)

    def cache_bb(self) -> BB:
        """Update and returns the bounding box of this shape"""
        _bb = cp.cpShapeCacheBB(self._shape)
        return BB(_bb.l, _bb.b, _bb.r, _bb.t)

    @property
    def bb(self) -> BB:
        """The bounding box :py:class:`BB` of the shape.

        Only guaranteed to be valid after :py:meth:`Shape.cache_bb` or
        :py:meth:`Space.step` is called. Moving a body that a shape is
        connected to does not update it's bounding box. For shapes used for
        queries that aren't attached to bodies, you can also use
        :py:meth:`Shape.update`.
        """
        _bb = cp.cpShapeGetBB(self._shape)
        return BB(_bb.l, _bb.b, _bb.r, _bb.t)

    def point_query(self, p: Tuple[float, float]) -> PointQueryInfo:
        """Check if the given point lies within the shape.

        A negative distance means the point is within the shape.

        :return: Tuple of (distance, info)
        :rtype: (float, :py:class:`PointQueryInfo`)
        """
        assert len(p) == 2
        info = ffi.new("cpPointQueryInfo *")
        _ = cp.cpShapePointQuery(self._shape, p, info)

        ud = int(ffi.cast("int", cp.cpShapeGetUserData(info.shape)))
        assert ud == self._id
        return PointQueryInfo(
            self,
            Vec2d(info.point.x, info.point.y),
            info.distance,
            Vec2d(info.gradient.x, info.gradient.y),
        )

    def segment_query(
        self, start: Tuple[float, float], end: Tuple[float, float], radius: float = 0
    ) -> SegmentQueryInfo:
        """Check if the line segment from start to end intersects the shape.

        :rtype: :py:class:`SegmentQueryInfo`
        """
        assert len(start) == 2
        assert len(end) == 2
        info = ffi.new("cpSegmentQueryInfo *")
        r = cp.cpShapeSegmentQuery(self._shape, start, end, radius, info)
        if r:
            ud = int(ffi.cast("int", cp.cpShapeGetUserData(info.shape)))
            assert ud == self._id
            return SegmentQueryInfo(
                self,
                Vec2d(info.point.x, info.point.y),
                Vec2d(info.normal.x, info.normal.y),
                info.alpha,
            )
        else:
            return SegmentQueryInfo(
                None,
                Vec2d(info.point.x, info.point.y),
                Vec2d(info.normal.x, info.normal.y),
                info.alpha,
            )

    def shapes_collide(self, b: "Shape") -> ContactPointSet:
        """Get contact information about this shape and shape b.

        :rtype: :py:class:`ContactPointSet`
        """
        _points = cp.cpShapesCollide(self._shape, b._shape)
        return ContactPointSet._from_cp(_points)

    @property
    def space(self) -> Optional["Space"]:
        """Get the :py:class:`Space` that shape has been added to (or
        None).
        """
        if self._space is not None:
            try:
                return self._space._get_self()  # ugly hack because of weakref
            except ReferenceError:
                return None
        else:
            return None

    def __getstate__(self) -> _State:
        """Return the state of this object

        This method allows the usage of the :mod:`copy` and :mod:`pickle`
        modules with this class.
        """
        d = super(Shape, self).__getstate__()

        if self.mass > 0:
            d["general"].append(("mass", self.mass))
        if self.density > 0:
            d["general"].append(("density", self.density))

        return d


class Circle(Shape):
    """A circle shape defined by a radius

    This is the fastest and simplest collision shape
    """

    _pickle_attrs_init = Shape._pickle_attrs_init + ["radius", "offset"]

    def __init__(
        self,
        body: Optional["Body"],
        radius: float,
        offset: Tuple[float, float] = (0, 0),
    ) -> None:
        """body is the body attach the circle to, offset is the offset from the
        body's center of gravity in body local coordinates.

        It is legal to send in None as body argument to indicate that this
        shape is not attached to a body. However, you must attach it to a body
        before adding the shape to a space or used for a space shape query.
        """
        assert len(offset) == 2
        body_body = ffi.NULL if body is None else body._body
        _shape = cp.cpCircleShapeNew(body_body, radius, offset)
        self._init(body, _shape)

    def unsafe_set_radius(self, r: float) -> None:
        """Unsafe set the radius of the circle.

        .. note::
            This change is only picked up as a change to the position
            of the shape's surface, but not it's velocity. Changing it will
            not result in realistic physical behavior. Only use if you know
            what you are doing!
        """
        cp.cpCircleShapeSetRadius(self._shape, r)

    @property
    def radius(self) -> float:
        """The Radius of the circle"""
        return cp.cpCircleShapeGetRadius(self._shape)

    def unsafe_set_offset(self, o: Tuple[float, float]) -> None:
        """Unsafe set the offset of the circle.

        .. note::
            This change is only picked up as a change to the position
            of the shape's surface, but not it's velocity. Changing it will
            not result in realistic physical behavior. Only use if you know
            what you are doing!
        """
        assert len(o) == 2
        cp.cpCircleShapeSetOffset(self._shape, o)

    @property
    def offset(self) -> Vec2d:
        """Offset. (body space coordinates)"""
        v = cp.cpCircleShapeGetOffset(self._shape)
        return Vec2d(v.x, v.y)


class Segment(Shape):
    """A line segment shape between two points

    Meant mainly as a static shape. Can be beveled in order to give them a
    thickness.
    """

    _pickle_attrs_init = Shape._pickle_attrs_init + ["a", "b", "radius"]

    def __init__(
        self,
        body: Optional["Body"],
        a: Tuple[float, float],
        b: Tuple[float, float],
        radius: float,
    ) -> None:
        """Create a Segment

        It is legal to send in None as body argument to indicate that this
        shape is not attached to a body. However, you must attach it to a body
        before adding the shape to a space or used for a space shape query.

        :param Body body: The body to attach the segment to
        :param a: The first endpoint of the segment
        :param b: The second endpoint of the segment
        :param float radius: The thickness of the segment
        """
        assert len(a) == 2
        assert len(b) == 2

        body_body = ffi.NULL if body is None else body._body
        _shape = cp.cpSegmentShapeNew(body_body, a, b, radius)
        self._init(body, _shape)

    def _get_a(self) -> Vec2d:
        v = cp.cpSegmentShapeGetA(self._shape)
        return Vec2d(v.x, v.y)

    a = property(_get_a, doc="""The first of the two endpoints for this segment""")

    def _get_b(self) -> Vec2d:
        v = cp.cpSegmentShapeGetB(self._shape)
        return Vec2d(v.x, v.y)

    b = property(_get_b, doc="""The second of the two endpoints for this segment""")

    def unsafe_set_endpoints(
        self, a: Tuple[float, float], b: Tuple[float, float]
    ) -> None:
        """Set the two endpoints for this segment

        .. note::
            This change is only picked up as a change to the position
            of the shape's surface, but not it's velocity. Changing it will
            not result in realistic physical behavior. Only use if you know
            what you are doing!
        """
        assert len(a) == 2
        assert len(b) == 2
        cp.cpSegmentShapeSetEndpoints(self._shape, a, b)

    @property
    def normal(self) -> Vec2d:
        """The normal"""
        v = cp.cpSegmentShapeGetNormal(self._shape)
        return Vec2d(v.x, v.y)

    def unsafe_set_radius(self, r: float) -> None:
        """Set the radius of the segment

        .. note::
            This change is only picked up as a change to the position
            of the shape's surface, but not it's velocity. Changing it will
            not result in realistic physical behavior. Only use if you know
            what you are doing!
        """
        cp.cpSegmentShapeSetRadius(self._shape, r)

    @property
    def radius(self) -> float:
        """The radius/thickness of the segment"""
        return cp.cpSegmentShapeGetRadius(self._shape)

    def set_neighbors(
        self, prev: Tuple[float, float], next: Tuple[float, float]
    ) -> None:
        """When you have a number of segment shapes that are all joined
        together, things can still collide with the "cracks" between the
        segments. By setting the neighbor segment endpoints you can tell
        Chipmunk to avoid colliding with the inner parts of the crack.
        """
        assert len(prev) == 2
        assert len(next) == 2
        cp.cpSegmentShapeSetNeighbors(self._shape, prev, next)


class Poly(Shape):
    """A convex polygon shape

    Slowest, but most flexible collision shape.
    """

    def __init__(
        self,
        body: Optional["Body"],
        vertices: Sequence[Tuple[float, float]],
        transform: Optional[Transform] = None,
        radius: float = 0,
    ) -> None:
        """Create a polygon.

        A convex hull will be calculated from the vertexes automatically.

        Adding a small radius will bevel the corners and can significantly
        reduce problems where the poly gets stuck on seams in your geometry.

        It is legal to send in None as body argument to indicate that this
        shape is not attached to a body. However, you must attach it to a body
        before adding the shape to a space or used for a space shape query.

        .. note::
            Make sure to put the vertices around (0,0) or the shape might
            behave strange.

            Either directly place the vertices like the below example:

            >>> import pymunk
            >>> w, h = 10, 20
            >>> vs = [(-w/2,-h/2), (w/2,-h/2), (w/2,h/2), (-w/2,h/2)]
            >>> poly_good = pymunk.Poly(None, vs)
            >>> print(poly_good.center_of_gravity)
            Vec2d(0.0, 0.0)

            Or use a transform to move them:

            >>> import pymunk
            >>> width, height = 10, 20
            >>> vs = [(0, 0), (width, 0), (width, height), (0, height)]
            >>> poly_bad = pymunk.Poly(None, vs)
            >>> print(poly_bad.center_of_gravity)
            Vec2d(5.0, 10.0)
            >>> t = pymunk.Transform(tx=-width/2, ty=-height/2)
            >>> poly_good = pymunk.Poly(None, vs, transform=t)
            >>> print(poly_good.center_of_gravity)
            Vec2d(0.0, 0.0)

        :param Body body: The body to attach the poly to
        :param [(float,float)] vertices: Define a convex hull of the polygon
            with a counterclockwise winding.
        :param Transform transform: Transform will be applied to every vertex.
        :param float radius: Set the radius of the poly shape

        """
        if transform is None:
            transform = Transform.identity()

        body_body = ffi.NULL if body is None else body._body
        _shape = cp.cpPolyShapeNew(
            body_body, len(vertices), vertices, transform, radius
        )
        self._init(body, _shape)

    def unsafe_set_radius(self, radius: float) -> None:
        """Unsafe set the radius of the poly.

        .. note::
            This change is only picked up as a change to the position
            of the shape's surface, but not it's velocity. Changing it will
            not result in realistic physical behavior. Only use if you know
            what you are doing!
        """
        cp.cpPolyShapeSetRadius(self._shape, radius)

    @property
    def radius(self) -> float:
        """The radius of the poly shape.

        Extends the poly in all directions with the given radius.
        """
        return cp.cpPolyShapeGetRadius(self._shape)

    @staticmethod
    def create_box(
        body: Optional["Body"], size: Tuple[float, float] = (10, 10), radius: float = 0
    ) -> "Poly":
        """Convenience function to create a box given a width and height.

        The boxes will always be centered at the center of gravity of the
        body you are attaching them to.  If you want to create an off-center
        box, you will need to use the normal constructor Poly(...).

        Adding a small radius will bevel the corners and can significantly
        reduce problems where the box gets stuck on seams in your geometry.

        :param Body body: The body to attach the poly to
        :param size: Size of the box as (width, height)
        :type size: (`float, float`)
        :param float radius: Radius of poly
        :rtype: :py:class:`Poly`
        """

        self = Poly.__new__(Poly)
        body_body = ffi.NULL if body is None else body._body
        _shape = cp.cpBoxShapeNew(body_body, size[0], size[1], radius)
        self._init(body, _shape)

        return self

    @staticmethod
    def create_box_bb(body: Optional["Body"], bb: BB, radius: float = 0) -> "Poly":
        """Convenience function to create a box shape from a :py:class:`BB`.

        The boxes will always be centered at the center of gravity of the
        body you are attaching them to.  If you want to create an off-center
        box, you will need to use the normal constructor Poly(..).

        Adding a small radius will bevel the corners and can significantly
        reduce problems where the box gets stuck on seams in your geometry.

        :param Body body: The body to attach the poly to
        :param BB bb: Size of the box
        :param float radius: Radius of poly
        :rtype: :py:class:`Poly`
        """

        self = Poly.__new__(Poly)
        body_body = ffi.NULL if body is None else body._body
        _shape = cp.cpBoxShapeNew2(body_body, bb, radius)
        self._init(body, _shape)

        return self

    def get_vertices(self) -> List[Vec2d]:
        """Get the vertices in local coordinates for the polygon

        If you need the vertices in world coordinates then the vertices can be
        transformed by adding the body position and each vertex rotated by the
        body rotation in the following way::

            >>> import pymunk
            >>> b = pymunk.Body()
            >>> b.position = 1,2
            >>> b.angle = 0.5
            >>> shape = pymunk.Poly(b, [(0,0), (10,0), (10,10)])
            >>> for v in shape.get_vertices():
            ...     x,y = v.rotated(shape.body.angle) + shape.body.position
            ...     (int(x), int(y))
            (1, 2)
            (9, 6)
            (4, 15)

        :return: The vertices in local coords
        :rtype: [:py:class:`Vec2d`]
        """
        verts = []
        l = cp.cpPolyShapeGetCount(self._shape)
        for i in range(l):
            v = cp.cpPolyShapeGetVert(self._shape, i)
            verts.append(Vec2d(v.x, v.y))
        return verts

    def unsafe_set_vertices(
        self,
        vertices: Sequence[Tuple[float, float]],
        transform: Optional[Transform] = None,
    ) -> None:
        """Unsafe set the vertices of the poly.

        .. note::
            This change is only picked up as a change to the position
            of the shape's surface, but not it's velocity. Changing it will
            not result in realistic physical behavior. Only use if you know
            what you are doing!
        """
        if transform is None:
            cp.cpPolyShapeSetVertsRaw(self._shape, len(vertices), vertices)
            return

        cp.cpPolyShapeSetVerts(self._shape, len(vertices), vertices, transform)

    def __getstate__(self) -> _State:
        """Return the state of this object

        This method allows the usage of the :mod:`copy` and :mod:`pickle`
        modules with this class.
        """
        d = super(Poly, self).__getstate__()

        d["init"].append(("vertices", self.get_vertices()))
        d["init"].append(("transform", None))
        d["init"].append(("radius", self.radius))
        return d



__docformat__ = "reStructuredText"


from typing import TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from .space import Space
    from .shapes import Shape

from ._chipmunk_cffi import ffi, lib
from .contact_point_set import ContactPointSet
from .vec2d import Vec2d


class Arbiter(object):
    """The Arbiter object encapsulates a pair of colliding shapes and all of
        the data about their collision.

        They are created when a collision starts, and persist until those
        shapes are no longer colliding.

    .. Warning::
        Because arbiters are handled by the space you should never
        hold onto a reference to an arbiter as you don't know when it will be
        destroyed! Use them within the callback where they are given to you
        and then forget about them or copy out the information you need from
        them.
    """

    def __init__(self, _arbiter: ffi.CData, space: "Space") -> None:
        """Initialize an Arbiter object from the Chipmunk equivalent struct
        and the Space.

        .. note::
            You should never need to create an instance of this class directly.
        """

        self._arbiter = _arbiter
        self._space = space

    def _get_contact_point_set(self) -> ContactPointSet:
        _set = lib.cpArbiterGetContactPointSet(self._arbiter)
        return ContactPointSet._from_cp(_set)

    def _set_contact_point_set(self, point_set: ContactPointSet) -> None:
        # This has to be done by fetching a new Chipmunk point set, update it
        # according to whats passed in and the pass that back to chipmunk due
        # to the fact that ContactPointSet doesnt contain a reference to the
        # corresponding c struct.
        _set = lib.cpArbiterGetContactPointSet(self._arbiter)
        _set.normal = point_set.normal

        if len(point_set.points) == _set.count:
            for i in range(_set.count):
                _set.points[i].pointA = point_set.points[0].point_a
                _set.points[i].pointB = point_set.points[0].point_b
                _set.points[i].distance = point_set.points[0].distance
        else:
            msg = "Expected {} points, got {} points in point_set".format(
                _set.count, len(point_set.points)
            )
            raise Exception(msg)

        lib.cpArbiterSetContactPointSet(self._arbiter, ffi.addressof(_set))

    contact_point_set = property(
        _get_contact_point_set,
        _set_contact_point_set,
        doc="""Contact point sets make getting contact information from the 
        Arbiter simpler.
        
        Return `ContactPointSet`""",
    )

    @property
    def shapes(self) -> Tuple["Shape", "Shape"]:
        """Get the shapes in the order that they were defined in the
        collision handler associated with this arbiter
        """
        shapeA_p = ffi.new("cpShape *[1]")
        shapeB_p = ffi.new("cpShape *[1]")

        lib.cpArbiterGetShapes(self._arbiter, shapeA_p, shapeB_p)

        a, b = self._space._get_shape(shapeA_p[0]), self._space._get_shape(shapeB_p[0])
        assert a is not None
        assert b is not None
        return a, b

    def _get_restitution(self) -> float:
        return lib.cpArbiterGetRestitution(self._arbiter)

    def _set_restitution(self, restitution: float) -> None:
        lib.cpArbiterSetRestitution(self._arbiter, restitution)

    restitution = property(
        _get_restitution,
        _set_restitution,
        doc="""The calculated restitution (elasticity) for this collision 
        pair. 
        
        Setting the value in a pre_solve() callback will override the value 
        calculated by the space. The default calculation multiplies the 
        elasticity of the two shapes together.
        """,
    )

    def _get_friction(self) -> float:
        return lib.cpArbiterGetFriction(self._arbiter)

    def _set_friction(self, friction: float) -> None:
        lib.cpArbiterSetFriction(self._arbiter, friction)

    friction = property(
        _get_friction,
        _set_friction,
        doc="""The calculated friction for this collision pair. 
        
        Setting the value in a pre_solve() callback will override the value 
        calculated by the space. The default calculation multiplies the 
        friction of the two shapes together.
        """,
    )

    def _get_surface_velocity(self) -> Vec2d:
        v = lib.cpArbiterGetSurfaceVelocity(self._arbiter)
        return Vec2d(v.x, v.y)

    def _set_surface_velocity(self, velocity: Vec2d) -> None:
        lib.cpArbiterSetSurfaceVelocity(self._arbiter, velocity)

    surface_velocity = property(
        _get_surface_velocity,
        _set_surface_velocity,
        doc="""The calculated surface velocity for this collision pair. 
        
        Setting the value in a pre_solve() callback will override the value 
        calculated by the space. the default calculation subtracts the 
        surface velocity of the second shape from the first and then projects 
        that onto the tangent of the collision. This is so that only 
        friction is affected by default calculation. Using a custom 
        calculation, you can make something that responds like a pinball 
        bumper, or where the surface velocity is dependent on the location 
        of the contact point.
        """,
    )

    @property
    def total_impulse(self) -> Vec2d:
        """Returns the impulse that was applied this step to resolve the
        collision.

        This property should only be called from a post-solve or each_arbiter
        callback.
        """
        v = lib.cpArbiterTotalImpulse(self._arbiter)
        return Vec2d(v.x, v.y)

    @property
    def total_ke(self) -> float:
        """The amount of energy lost in a collision including static, but
        not dynamic friction.

        This property should only be called from a post-solve or each_arbiter callback.
        """
        return lib.cpArbiterTotalKE(self._arbiter)

    @property
    def is_first_contact(self) -> bool:
        """Returns true if this is the first step the two shapes started
        touching.

        This can be useful for sound effects for instance. If its the first
        frame for a certain collision, check the energy of the collision in a
        post_step() callback and use that to determine the volume of a sound
        effect to play.
        """
        return bool(lib.cpArbiterIsFirstContact(self._arbiter))

    @property
    def is_removal(self) -> bool:
        """Returns True during a separate() callback if the callback was
        invoked due to an object removal.
        """
        return bool(lib.cpArbiterIsRemoval(self._arbiter))

    @property
    def normal(self) -> Vec2d:
        """Returns the normal of the collision."""
        v = lib.cpArbiterGetNormal(self._arbiter)
        return Vec2d(v.x, v.y)



import pickle
import unittest
from typing import Any

import pymunk as p
from pymunk.vec2d import Vec2d

####################################################################


class UnitTestShape(unittest.TestCase):
    def testId(self) -> None:
        c = p.Circle(None, 4)
        self.assertGreater(c._id, 0)

    def testPointQuery(self) -> None:
        b = p.Body(10, 10)
        c = p.Circle(b, 5)
        c.cache_bb()

        info = c.point_query((0, 0))

        self.assertEqual(info.shape, c)
        self.assertEqual(info.point, (0, 0))
        self.assertEqual(info.distance, -5)
        self.assertEqual(info.gradient, (0, 1))

        info = c.point_query((11, 0))

        self.assertEqual(info.shape, c)
        self.assertEqual(info.point, (5, 0))
        self.assertEqual(info.distance, 6)
        self.assertEqual(info.gradient, (1, 0))

    def testSegmentQuery(self) -> None:
        s = p.Space()
        b = p.Body(10, 10)
        c = p.Circle(b, 5)
        c.cache_bb()

        info = c.segment_query((10, -50), (10, 50))
        self.assertEqual(info.shape, None)
        self.assertEqual(info.point, (10, 50))
        self.assertEqual(info.normal, (0, 0))
        self.assertEqual(info.alpha, 1.0)

        info = c.segment_query((10, -50), (10, 50), 6)
        self.assertEqual(info.shape, c)

        info = c.segment_query((0, -50), (0, 50))
        self.assertEqual(info.shape, c)
        self.assertAlmostEqual(info.point.x, 0)
        self.assertAlmostEqual(info.point.y, -5)
        self.assertAlmostEqual(info.normal.x, 0)
        self.assertAlmostEqual(info.normal.y, -1)
        self.assertEqual(info.alpha, 0.45)

    def testMass(self) -> None:
        c = p.Circle(None, 1)
        self.assertEqual(c.mass, 0)
        c.mass = 2
        self.assertEqual(c.mass, 2)

    def testDensity(self) -> None:
        c = p.Circle(None, 1)
        self.assertEqual(c.density, 0)
        c.density = 2
        self.assertEqual(c.density, 2)

    def testMoment(self) -> None:
        c = p.Circle(None, 5)
        self.assertEqual(c.moment, 0)
        c.density = 2
        self.assertAlmostEqual(c.moment, 1963.4954084936207)
        c.density = 0
        c.mass = 2
        self.assertAlmostEqual(c.moment, 25)

    def testArea(self) -> None:
        c = p.Circle(None, 5)
        self.assertEqual(c.area, 78.53981633974483)

    def testCenterOfGravity(self) -> None:
        c = p.Circle(None, 5)
        self.assertEqual(c.center_of_gravity, (0, 0))
        c = p.Circle(None, 5, (10, 5))
        self.assertEqual(c.center_of_gravity.x, 10)
        self.assertEqual(c.center_of_gravity.y, 5)

    def testNoBody(self) -> None:
        c = p.Circle(None, 1)
        self.assertEqual(c.body, None)

    def testRemoveBody(self) -> None:
        b = p.Body(1, 1)
        c = p.Circle(b, 1)
        c.body = None

        self.assertEqual(c.body, None)
        self.assertEqual(len(b.shapes), 0)

    def testSwitchBody(self) -> None:
        b1 = p.Body(1, 1)
        b2 = p.Body(1, 1)
        c = p.Circle(b1, 1)
        self.assertEqual(c.body, b1)
        self.assertTrue(c in b1.shapes)
        self.assertTrue(c not in b2.shapes)
        c.body = b2
        self.assertEqual(c.body, b2)
        self.assertTrue(c not in b1.shapes)
        self.assertTrue(c in b2.shapes)

    def testSensor(self) -> None:
        b1 = p.Body(1, 1)
        c = p.Circle(b1, 1)
        self.assertFalse(c.sensor)
        c.sensor = True
        self.assertTrue(c.sensor)

    def testElasticity(self) -> None:
        b1 = p.Body(1, 1)
        c = p.Circle(b1, 1)
        self.assertEqual(c.elasticity, 0)
        c.elasticity = 1
        self.assertEqual(c.elasticity, 1)

    def testFriction(self) -> None:
        b1 = p.Body(1, 1)
        c = p.Circle(b1, 1)
        self.assertEqual(c.friction, 0)
        c.friction = 1
        self.assertEqual(c.friction, 1)

    def testSurfaceVelocity(self) -> None:
        b1 = p.Body(1, 1)
        c = p.Circle(b1, 1)
        self.assertEqual(c.surface_velocity, (0, 0))
        c.surface_velocity = (1, 2)
        self.assertEqual(c.surface_velocity, (1, 2))

    def testCollisionType(self) -> None:
        b1 = p.Body(1, 1)
        c = p.Circle(b1, 1)
        self.assertEqual(c.collision_type, 0)
        c.collision_type = 1
        self.assertEqual(c.collision_type, 1)

    def testFilter(self) -> None:
        b1 = p.Body(1, 1)
        c = p.Circle(b1, 1)
        self.assertEqual(c.filter, p.ShapeFilter(0, 0xFFFFFFFF, 0xFFFFFFFF))
        c.filter = p.ShapeFilter(1, 0xFFFFFFF2, 0xFFFFFFF3)
        self.assertEqual(c.filter, p.ShapeFilter(1, 0xFFFFFFF2, 0xFFFFFFF3))

    def testSpace(self) -> None:
        b1 = p.Body(1, 1)
        c = p.Circle(b1, 1)
        self.assertEqual(c.space, None)
        s = p.Space()
        s.add(b1, c)
        self.assertEqual(c.space, s)

    def testShapesCollide(self) -> None:
        b1 = p.Body(1, 1)
        s1 = p.Circle(b1, 10)

        b2 = p.Body(1, 1)
        b2.position = 30, 30
        s2 = p.Circle(b2, 10)

        c = s1.shapes_collide(s2)
        self.assertEqual(c.normal, (1, 0))
        self.assertEqual(len(c.points), 1)
        point = c.points[0]
        self.assertEqual(point.point_a, (10, 0))
        self.assertEqual(point.point_b, (-10, 0))
        self.assertEqual(point.distance, -20)

    def testPickle(self) -> None:
        b = p.Body(1, 2)
        c = p.Circle(b, 3, (4, 5))
        c.sensor = True
        c.collision_type = 6
        c.filter = p.ShapeFilter()
        c.elasticity = 7
        c.friction = 8
        c.surface_velocity = (9, 10)

        s = pickle.dumps(c)
        c2 = pickle.loads(s)

        self.assertEqual(c.sensor, c2.sensor)
        self.assertEqual(c.collision_type, c2.collision_type)
        self.assertEqual(c.filter, c2.filter)
        self.assertEqual(c.elasticity, c2.elasticity)
        self.assertEqual(c.friction, c2.friction)
        self.assertEqual(c.surface_velocity, c2.surface_velocity)
        self.assertEqual(c.density, c2.density)
        self.assertEqual(c.mass, c2.mass)
        self.assertEqual(c.body.mass, c2.body.mass)

        c = p.Circle(None, 1)
        c.density = 3

        s = pickle.dumps(c)
        c2 = pickle.loads(s)

        self.assertEqual(c.mass, c2.mass)
        self.assertEqual(c.density, c2.density)

        c2 = c.copy()


class UnitTestCircle(unittest.TestCase):
    def testCircleBB(self) -> None:
        b = p.Body(10, 10)
        c = p.Circle(b, 5)

        c.cache_bb()

        self.assertEqual(c.bb, p.BB(-5.0, -5.0, 5.0, 5.0))

    def testCircleNoBody(self) -> None:
        c = p.Circle(None, 5)

        bb = c.update(p.Transform(1, 2, 3, 4, 5, 6))
        self.assertEqual(c.bb, bb)
        self.assertEqual(c.bb, p.BB(0, 1, 10, 11))

    def testOffset(self) -> None:
        c = p.Circle(None, 5, (1, 2))

        self.assertEqual(c.offset, (1, 2))

    def testOffsetUnsafe(self) -> None:
        c = p.Circle(None, 5, (1, 2))

        c.unsafe_set_offset((3, 4))

        self.assertEqual(c.offset, (3, 4))

    def testRadius(self) -> None:
        c = p.Circle(None, 5)

        self.assertEqual(c.radius, 5)

    def testRadiusUnsafe(self) -> None:
        c = p.Circle(None, 5)

        c.unsafe_set_radius(3)

        self.assertEqual(c.radius, 3)

    def testPickle(self) -> None:
        c = p.Circle(None, 3, (4, 5))

        s = pickle.dumps(c)
        c2 = pickle.loads(s)

        self.assertEqual(c.radius, c2.radius)
        self.assertEqual(c.offset, c2.offset)


class UnitTestSegment(unittest.TestCase):
    def testBB(self) -> None:
        s = p.Space()
        b = p.Body(10, 10)
        c = p.Segment(b, (2, 2), (2, 3), 2)

        c.cache_bb()

        self.assertEqual(c.bb, p.BB(0, 0, 4.0, 5.0))

    def testProperties(self) -> None:
        c = p.Segment(None, (2, 2), (2, 3), 4)

        self.assertEqual(c.a, (2, 2))
        self.assertEqual(c.b, (2, 3))
        self.assertEqual(c.normal, (1, 0))
        self.assertEqual(c.radius, 4)

    def testPropertiesUnsafe(self) -> None:
        c = p.Segment(None, (2, 2), (2, 3), 4)

        c.unsafe_set_endpoints((3, 4), (5, 6))
        self.assertEqual(c.a, (3, 4))
        self.assertEqual(c.b, (5, 6))

        c.unsafe_set_radius(5)
        self.assertEqual(c.radius, 5)

    def testSetNeighbors(self) -> None:
        c = p.Segment(None, (2, 2), (2, 3), 1)
        c.set_neighbors((2, 2), (2, 3))

    def testSegmentSegmentCollision(self) -> None:
        s = p.Space()
        b1 = p.Body(10, 10)
        c1 = p.Segment(b1, (-1, -1), (1, 1), 1)
        b2 = p.Body(10, 10)
        c2 = p.Segment(b2, (1, -1), (-1, 1), 1)

        s.add(b1, b2, c1, c2)

        self.num_of_begins = 0

        def begin(arb: p.Arbiter, space: p.Space, data: Any) -> bool:
            self.num_of_begins += 1
            return True

        s.add_default_collision_handler().begin = begin
        s.step(0.1)

        self.assertEqual(1, self.num_of_begins)

    def testPickle(self) -> None:
        c = p.Segment(None, (1, 2), (3, 4), 5)

        s = pickle.dumps(c)
        c2 = pickle.loads(s)

        self.assertEqual(c.a, c2.a)
        self.assertEqual(c.b, c2.b)
        self.assertEqual(c.radius, c2.radius)


class UnitTestPoly(unittest.TestCase):
    def testInit(self) -> None:
        c = p.Poly(None, [(0, 0), (10, 10), (20, 0), (-10, 10)], None, 0)

        b = p.Body(1, 2)
        c = p.Poly(b, [(0, 0), (10, 10), (20, 0), (-10, 10)], p.Transform.identity(), 6)

    def testVertices(self) -> None:
        vs = [(-10, 10), (0, 0), (20, 0), (10, 10)]
        c = p.Poly(None, vs, None, 0)

        self.assertEqual(c.get_vertices(), vs)

        c = p.Poly(None, vs, p.Transform(1, 2, 3, 4, 5, 6), 0)

        vs2 = [(5.0, 6.0), (25.0, 26.0), (45.0, 66.0), (25.0, 46.0)]
        self.assertEqual(c.get_vertices(), vs2)

    def testVerticesUnsafe(self) -> None:
        vs = [(-10, 10), (0, 0), (20, 0), (10, 10)]
        c = p.Poly(None, vs, None, 0)

        vs2 = [(-3, 3), (0, 0), (3, 0)]
        c.unsafe_set_vertices(vs2)
        self.assertEqual(c.get_vertices(), vs2)

        vs3 = [(-4, 4), (0, 0), (4, 0)]
        c.unsafe_set_vertices(vs3, p.Transform.identity())
        self.assertEqual(c.get_vertices(), vs3)

    def testBB(self) -> None:
        c = p.Poly(None, [(2, 2), (4, 3), (3, 5)])
        bb = c.update(p.Transform.identity())
        self.assertEqual(bb, c.bb)
        self.assertEqual(c.bb, p.BB(2, 2, 4, 5))

        b = p.Body(1, 2)
        c = p.Poly(b, [(2, 2), (4, 3), (3, 5)])
        c.cache_bb()
        self.assertEqual(c.bb, p.BB(2, 2, 4, 5))

        s = p.Space()
        b = p.Body(1, 2)
        c = p.Poly(b, [(2, 2), (4, 3), (3, 5)])
        s.add(b, c)
        self.assertEqual(c.bb, p.BB(2, 2, 4, 5))

    def testRadius(self) -> None:
        c = p.Poly(None, [(2, 2), (4, 3), (3, 5)], radius=10)
        self.assertEqual(c.radius, 10)

    def testRadiusUnsafe(self) -> None:
        c = p.Poly(None, [(2, 2), (4, 3), (3, 5)], radius=10)

        c.unsafe_set_radius(20)

        self.assertEqual(c.radius, 20)

    def testCreateBox(self) -> None:
        c = p.Poly.create_box(None, (4, 2), 3)
        self.assertEqual(c.get_vertices(), [(2, -1), (2, 1), (-2, 1), (-2, -1)])

        c = p.Poly.create_box_bb(None, p.BB(1, 2, 3, 4), 3)
        self.assertEqual(c.get_vertices(), [(3, 2), (3, 4), (1, 4), (1, 2)])

    def testPickle(self) -> None:
        c = p.Poly(None, [(1, 2), (3, 4), (5, 6)], radius=5)

        s = pickle.dumps(c)
        c2 = pickle.loads(s)

        self.assertEqual(c.get_vertices(), c2.get_vertices())
        self.assertEqual(c.radius, c2.radius)


####################################################################
if __name__ == "__main__":
    print("testing pymunk version " + p.version)
    unittest.main()



__docformat__ = "reStructuredText"

from typing import TYPE_CHECKING, ClassVar, NamedTuple, Optional, Sequence, Tuple, Type

if TYPE_CHECKING:
    from .shapes import Shape
    from types import TracebackType

import math

from ._chipmunk_cffi import ffi, lib
from .body import Body
from .transform import Transform
from .vec2d import Vec2d

_DrawFlags = int


class SpaceDebugColor(NamedTuple):
    """Color tuple used by the debug drawing API."""

    r: float
    g: float
    b: float
    a: float

    def as_int(self) -> Tuple[int, int, int, int]:
        """Return the color as a tuple of ints, where each value is rounded.

        >>> SpaceDebugColor(0, 51.1, 101.9, 255).as_int()
        (0, 51, 102, 255)
        """
        return round(self[0]), round(self[1]), round(self[2]), round(self[3])

    def as_float(self) -> Tuple[float, float, float, float]:
        """Return the color as a tuple of floats, each value divided by 255.

        >>> SpaceDebugColor(0, 51, 102, 255).as_float()
        (0.0, 0.2, 0.4, 1.0)
        """
        return self[0] / 255.0, self[1] / 255.0, self[2] / 255.0, self[3] / 255.0


class SpaceDebugDrawOptions(object):
    """SpaceDebugDrawOptions configures debug drawing.

    If appropriate its usually easy to use the supplied draw implementations
    directly: pymunk.pygame_util, pymunk.pyglet_util and pymunk.matplotlib_util.
    """

    DRAW_SHAPES: ClassVar[_DrawFlags] = lib.CP_SPACE_DEBUG_DRAW_SHAPES
    """Draw shapes.  
    
    Use on the flags property to control if shapes should be drawn or not.
    """

    DRAW_CONSTRAINTS: ClassVar[_DrawFlags] = lib.CP_SPACE_DEBUG_DRAW_CONSTRAINTS
    """Draw constraints. 
    
    Use on the flags property to control if constraints should be drawn or not.
    """

    DRAW_COLLISION_POINTS: ClassVar[
        _DrawFlags
    ] = lib.CP_SPACE_DEBUG_DRAW_COLLISION_POINTS
    """Draw collision points.
    
    Use on the flags property to control if collision points should be drawn or
    not.
    """

    shape_dynamic_color: SpaceDebugColor = SpaceDebugColor(52, 152, 219, 255)
    shape_static_color: SpaceDebugColor = SpaceDebugColor(149, 165, 166, 255)
    shape_kinematic_color: SpaceDebugColor = SpaceDebugColor(39, 174, 96, 255)
    shape_sleeping_color: SpaceDebugColor = SpaceDebugColor(114, 148, 168, 255)

    def __init__(self) -> None:
        _options = ffi.new("cpSpaceDebugDrawOptions *")
        self._options = _options
        self._options.transform = Transform.identity()
        self.shape_outline_color = SpaceDebugColor(44, 62, 80, 255)
        self.constraint_color = SpaceDebugColor(142, 68, 173, 255)
        self.collision_point_color = SpaceDebugColor(231, 76, 60, 255)

        # Set to false to bypass chipmunk shape drawing code
        self._use_chipmunk_debug_draw = True

        @ffi.callback("cpSpaceDebugDrawCircleImpl")
        def f1(pos, angle, radius, outline_color, fill_color, _):  # type: ignore
            self.draw_circle(
                Vec2d(pos.x, pos.y),
                angle,
                radius,
                self._c(outline_color),
                self._c(fill_color),
            )

        _options.drawCircle = f1

        @ffi.callback("cpSpaceDebugDrawSegmentImpl")
        def f2(a, b, color, _):  # type: ignore
            # sometimes a and/or b can be nan. For example if both endpoints
            # of a spring is at the same position. In those cases skip calling
            # the drawing method.
            if math.isnan(a.x) or math.isnan(a.y) or math.isnan(b.x) or math.isnan(b.y):
                return
            self.draw_segment(
                Vec2d(a.x, a.y),
                Vec2d(b.x, b.y),
                self._c(color),
            )

        _options.drawSegment = f2

        @ffi.callback("cpSpaceDebugDrawFatSegmentImpl")
        def f3(a, b, radius, outline_color, fill_color, _):  # type: ignore
            self.draw_fat_segment(
                Vec2d(a.x, a.y),
                Vec2d(b.x, b.y),
                radius,
                self._c(outline_color),
                self._c(fill_color),
            )

        _options.drawFatSegment = f3

        @ffi.callback("cpSpaceDebugDrawPolygonImpl")
        def f4(count, verts, radius, outline_color, fill_color, _):  # type: ignore
            vs = []
            for i in range(count):
                vs.append(Vec2d(verts[i].x, verts[i].y))
            self.draw_polygon(vs, radius, self._c(outline_color), self._c(fill_color))

        _options.drawPolygon = f4

        @ffi.callback("cpSpaceDebugDrawDotImpl")
        def f5(size, pos, color, _):  # type: ignore
            self.draw_dot(size, Vec2d(pos.x, pos.y), self._c(color))

        _options.drawDot = f5

        @ffi.callback("cpSpaceDebugDrawColorForShapeImpl")
        def f6(_shape, data):  # type: ignore
            space = ffi.from_handle(data)
            shape = space._get_shape(_shape)
            return self.color_for_shape(shape)

        _options.colorForShape = f6

        self.flags = (
            SpaceDebugDrawOptions.DRAW_SHAPES
            | SpaceDebugDrawOptions.DRAW_CONSTRAINTS
            | SpaceDebugDrawOptions.DRAW_COLLISION_POINTS
        )

        self._callbacks = [f1, f2, f3, f4, f5, f6]

    def _get_shape_outline_color(self) -> SpaceDebugColor:
        return self._c(self._options.shapeOutlineColor)

    def _set_shape_outline_color(self, c: SpaceDebugColor) -> None:
        self._options.shapeOutlineColor = c

    shape_outline_color = property(
        _get_shape_outline_color,
        _set_shape_outline_color,
        doc="""The outline color of shapes.
        
        Should be a tuple of 4 ints between 0 and 255 (r,g,b,a).

        Example:

        >>> import pymunk
        >>> s = pymunk.Space()
        >>> c = pymunk.Circle(s.static_body, 10)
        >>> s.add(c)
        >>> options = pymunk.SpaceDebugDrawOptions()
        >>> s.debug_draw(options)
        draw_circle (Vec2d(0.0, 0.0), 0.0, 10.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=149.0, g=165.0, b=166.0, a=255.0))
        >>> options.shape_outline_color = (10,20,30,40)
        >>> s.debug_draw(options)
        draw_circle (Vec2d(0.0, 0.0), 0.0, 10.0, SpaceDebugColor(r=10.0, g=20.0, b=30.0, a=40.0), SpaceDebugColor(r=149.0, g=165.0, b=166.0, a=255.0))

        """,
    )

    def _get_constraint_color(self) -> SpaceDebugColor:
        return self._c(self._options.constraintColor)

    def _set_constraint_color(self, c: SpaceDebugColor) -> None:
        self._options.constraintColor = c

    constraint_color = property(
        _get_constraint_color,
        _set_constraint_color,
        doc="""The color of constraints.

        Should be a tuple of 4 ints between 0 and 255 (r,g,b,a).
        
        Example:

        >>> import pymunk
        >>> s = pymunk.Space()
        >>> b = pymunk.Body(1, 10)
        >>> j = pymunk.PivotJoint(s.static_body, b, (0,0))
        >>> s.add(j)
        >>> options = pymunk.SpaceDebugDrawOptions()
        >>> s.debug_draw(options)
        draw_dot (5.0, Vec2d(0.0, 0.0), SpaceDebugColor(r=142.0, g=68.0, b=173.0, a=255.0))
        draw_dot (5.0, Vec2d(0.0, 0.0), SpaceDebugColor(r=142.0, g=68.0, b=173.0, a=255.0))
        >>> options.constraint_color = (10,20,30,40)
        >>> s.debug_draw(options)
        draw_dot (5.0, Vec2d(0.0, 0.0), SpaceDebugColor(r=10.0, g=20.0, b=30.0, a=40.0))
        draw_dot (5.0, Vec2d(0.0, 0.0), SpaceDebugColor(r=10.0, g=20.0, b=30.0, a=40.0))

        """,
    )

    def _get_collision_point_color(self) -> SpaceDebugColor:
        return self._c(self._options.collisionPointColor)

    def _set_collision_point_color(self, c: SpaceDebugColor) -> None:
        self._options.collisionPointColor = c

    collision_point_color = property(
        _get_collision_point_color,
        _set_collision_point_color,
        doc="""The color of collisions.

        Should be a tuple of 4 ints between 0 and 255 (r,g,b,a).
        
        Example:

        >>> import pymunk
        >>> s = pymunk.Space()
        >>> b = pymunk.Body(1,10)
        >>> c1 = pymunk.Circle(b, 10)
        >>> c2 = pymunk.Circle(s.static_body, 10)
        >>> s.add(b, c1, c2)
        >>> s.step(1)
        >>> options = pymunk.SpaceDebugDrawOptions()
        >>> s.debug_draw(options)
        draw_circle (Vec2d(0.0, 0.0), 0.0, 10.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=52.0, g=152.0, b=219.0, a=255.0))
        draw_circle (Vec2d(0.0, 0.0), 0.0, 10.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=149.0, g=165.0, b=166.0, a=255.0))
        draw_segment (Vec2d(8.0, 0.0), Vec2d(-8.0, 0.0), SpaceDebugColor(r=231.0, g=76.0, b=60.0, a=255.0))
        >>> options.collision_point_color = (10,20,30,40)
        >>> s.debug_draw(options)
        draw_circle (Vec2d(0.0, 0.0), 0.0, 10.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=52.0, g=152.0, b=219.0, a=255.0))
        draw_circle (Vec2d(0.0, 0.0), 0.0, 10.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=149.0, g=165.0, b=166.0, a=255.0))
        draw_segment (Vec2d(8.0, 0.0), Vec2d(-8.0, 0.0), SpaceDebugColor(r=10.0, g=20.0, b=30.0, a=40.0))
        """,
    )

    def __enter__(self) -> None:
        pass

    def __exit__(
        self,
        type: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional["TracebackType"],
    ) -> None:
        pass

    def _c(self, color: ffi.CData) -> SpaceDebugColor:
        return SpaceDebugColor(color.r, color.g, color.b, color.a)

    def _get_flags(self) -> _DrawFlags:
        return self._options.flags

    def _set_flags(self, f: _DrawFlags) -> None:
        self._options.flags = f

    flags = property(
        _get_flags,
        _set_flags,
        doc="""Bit flags which of shapes, joints and collisions should be drawn.

        By default all 3 flags are set, meaning shapes, joints and collisions 
        will be drawn.

        Example using the basic text only DebugDraw implementation (normally
        you would the desired backend instead, such as 
        `pygame_util.DrawOptions` or `pyglet_util.DrawOptions`):

        >>> import pymunk
        >>> s = pymunk.Space()
        >>> b = pymunk.Body()
        >>> c = pymunk.Circle(b, 10)
        >>> c.mass = 3
        >>> s.add(b, c)
        >>> s.add(pymunk.Circle(s.static_body, 3))
        >>> s.step(0.01)
        >>> options = pymunk.SpaceDebugDrawOptions() 
        
        >>> # Only draw the shapes, nothing else:
        >>> options.flags = pymunk.SpaceDebugDrawOptions.DRAW_SHAPES
        >>> s.debug_draw(options) 
        draw_circle (Vec2d(0.0, 0.0), 0.0, 10.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=52.0, g=152.0, b=219.0, a=255.0))
        draw_circle (Vec2d(0.0, 0.0), 0.0, 3.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=149.0, g=165.0, b=166.0, a=255.0))

        >>> # Draw the shapes and collision points:
        >>> options.flags = pymunk.SpaceDebugDrawOptions.DRAW_SHAPES
        >>> options.flags |= pymunk.SpaceDebugDrawOptions.DRAW_COLLISION_POINTS
        >>> s.debug_draw(options)
        draw_circle (Vec2d(0.0, 0.0), 0.0, 10.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=52.0, g=152.0, b=219.0, a=255.0))
        draw_circle (Vec2d(0.0, 0.0), 0.0, 3.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=149.0, g=165.0, b=166.0, a=255.0))
        draw_segment (Vec2d(1.0, 0.0), Vec2d(-8.0, 0.0), SpaceDebugColor(r=231.0, g=76.0, b=60.0, a=255.0))
        
        """,
    )

    def _get_transform(self) -> Transform:
        t = self._options.transform
        return Transform(t.a, t.b, t.c, t.d, t.tx, t.ty)

    def _set_transform(self, t: Transform) -> None:
        self._options.transform = t

    transform = property(
        _get_transform,
        _set_transform,
        doc="""The transform is applied before drawing, e.g for scaling or 
        translation.

        Example: 

        >>> import pymunk
        >>> s = pymunk.Space()
        >>> c = pymunk.Circle(s.static_body, 10)
        >>> s.add(c)
        >>> options = pymunk.SpaceDebugDrawOptions() 
        >>> s.debug_draw(options) 
        draw_circle (Vec2d(0.0, 0.0), 0.0, 10.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=149.0, g=165.0, b=166.0, a=255.0))
        >>> options.transform = pymunk.Transform.scaling(2)
        >>> s.debug_draw(options)
        draw_circle (Vec2d(0.0, 0.0), 0.0, 20.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=149.0, g=165.0, b=166.0, a=255.0))
        >>> options.transform = pymunk.Transform.translation(2,3)
        >>> s.debug_draw(options)
        draw_circle (Vec2d(2.0, 3.0), 0.0, 10.0, SpaceDebugColor(r=44.0, g=62.0, b=80.0, a=255.0), SpaceDebugColor(r=149.0, g=165.0, b=166.0, a=255.0))
        
        .. Note::
            Not all tranformations are supported by the debug drawing logic. 
            Uniform scaling and translation are supported, but not rotation,
            linear stretching or shearing. 
        """,
    )

    def draw_circle(
        self,
        pos: Vec2d,
        angle: float,
        radius: float,
        outline_color: SpaceDebugColor,
        fill_color: SpaceDebugColor,
    ) -> None:
        print("draw_circle", (pos, angle, radius, outline_color, fill_color))

    def draw_segment(self, a: Vec2d, b: Vec2d, color: SpaceDebugColor) -> None:
        print("draw_segment", (a, b, color))

    def draw_fat_segment(
        self,
        a: Vec2d,
        b: Vec2d,
        radius: float,
        outline_color: SpaceDebugColor,
        fill_color: SpaceDebugColor,
    ) -> None:
        print("draw_fat_segment", (a, b, radius, outline_color, fill_color))

    def draw_polygon(
        self,
        verts: Sequence[Vec2d],
        radius: float,
        outline_color: SpaceDebugColor,
        fill_color: SpaceDebugColor,
    ) -> None:
        print("draw_polygon", (verts, radius, outline_color, fill_color))

    def draw_dot(self, size: float, pos: Vec2d, color: SpaceDebugColor) -> None:
        print("draw_dot", (size, pos, color))

    def draw_shape(self, shape: "Shape") -> None:
        print("draw_shape", shape)

    def color_for_shape(self, shape: "Shape") -> SpaceDebugColor:
        if hasattr(shape, "color"):
            return SpaceDebugColor(*shape.color)  # type: ignore

        color = self.shape_dynamic_color
        if shape.body != None:
            if shape.body.body_type == Body.STATIC:
                color = self.shape_static_color
            elif shape.body.body_type == Body.KINEMATIC:
                color = self.shape_kinematic_color
            elif shape.body.is_sleeping:
                color = self.shape_sleeping_color

        return color
